# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pluggy
import click

from git_system_follower.plugins.cli.packages import NAME
from git_system_follower.typings.cli import PackageCLISource, PackageCLITarGz, PackageCLIImage


__all__ = ['HookSpec']


hookspec = pluggy.HookspecMarker(NAME)


class HookSpec:
    value: str | None
    gears: list[PackageCLISource | PackageCLITarGz | PackageCLIImage] | None

    def __init__(self):
        self.value = None

    def process(self, value, **kwargs) -> bool:
        match = self.match(value)
        if match:
            self.value = value
            self.gears = self.get_gears(value, **kwargs)
        return match

    def __str__(self) -> str:
        gears = [str(gear) for gear in self.gears]
        msg = ', '.join(gears) if gears else 'No Gears inside'
        return f'{self.value} ({msg})'

    @hookspec
    def plugin_options(self) -> list[click.option]:
        """ Get plugin options for CLI options """
        return []

    @hookspec
    def match(self, value: str) -> bool:
        """ Is this value a package of this type

        :param value: GSF cli argument
        :return: true/false
        """
        return False

    @hookspec
    def get_gears(self, value: str, **kwargs) -> list[PackageCLISource | PackageCLITarGz | PackageCLIImage]:
        """ Get Gears from this CLI argument type """
        return []



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with custom errors """


class CLIParamsError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class ParsePackageNameError(CLIParamsError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class RemoteRepositoryError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class InstallationError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class UninstallationError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


""" --------------------------------------- """


class PackageNotFoundError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class MaxDependencyDepthError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageDescriptionFileError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class DescriptionSectionError(PackageDescriptionFileError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


""" --------------------------------------- """


class PackageNamePolicyError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageApiError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageInterfaceError(PackageApiError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageAPIDevelopmentError(PackageInterfaceError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageCICDVariablePolicyError(PackageInterfaceError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PackageTemplatePolicyError(PackageInterfaceError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class StateFileError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class HashesMismatch(StateFileError):
    def __init__(self, msg, state_file_hash: str, generated_hash: str, *args):
        self.state_file_hash = state_file_hash
        self.generated_hash = generated_hash
        super().__init__(msg, *args)


class DownloadPackageError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class UnknownRegistryError(DownloadPackageError):
    """ Raised when the registry type cannot be determined """
    def __init__(self, msg, *args):
        super().__init__(msg, *args)

""" --------------------------------------- """

class PluginError(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class InvalidPlugin(PluginError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


class PluginExecutionError(PluginError):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api to work with CI/CD variables for package developers """
import json

from gitlab.base import RESTObject

from git_system_follower.variables import PACKAGE_API_RESULT as __PACKAGE_API_RESULT
from git_system_follower.develop.api.types import Parameters
from git_system_follower.errors import PackageCICDVariablePolicyError
from git_system_follower.package.cicd_variables import (
    CICDVariable,
    create_variable as __create_variable,
    delete_variable as __delete_variable
)


__all__ = ['CICDVariable', 'create_variable', 'delete_variable']


def create_variable(parameters: Parameters, variable: CICDVariable, *, is_force: bool = False) -> RESTObject | None:
    """ Create CI/CD variable using gitlab REST API

    If <is_force> parameter is False, then it will necessarily be safe to create CI/CD variable:
        1. If CI/CD variable doesn't exist: create CI/CD variable
        2. If CI/CD variable exists:
            1. CI/CD variables values doesn't match: notification of this (warning)
            2. CI/CD variables values matches: notification of this (info)

    If <is_force> parameter is True, then it will necessarily be force to create CI/CD variable:
        1. If CI/CD variable doesn't exist: create CI/CD variable
        2. If CI/CD variable exists:
            1. CI/CD variables values doesn't match: overwrite this CI/CD variable, notification of this (warning)
            2. CI/CD variables values content matches: notification of this (info)

    :param parameters: parameters that were passed to the package api
    :param variable: CI/CD variable to be created
    :param is_force: forced creation (ignore variable value)

    :return: creation response if variable is created
    """
    system_params = parameters._Parameters__system_params
    if variable['name'] in system_params.created_cicd_vars_names:
        raise PackageCICDVariablePolicyError(f"{variable['name']} CI/CD variable used in another package. According "
                                             f"to package manager policy, you cannot use the same variables in "
                                             f"different packages")

    is_force = True if is_force or system_params.is_force else False
    response = __create_variable(
        system_params.project, variable, is_force=is_force
    )
    __add_info_about_variable(variable)
    return response


def delete_variable(parameters: Parameters, variable: CICDVariable, *, is_force: bool = False) -> None:
    """ Delete CI/CD variable using gitlab REST API

    If <is_force> parameter is False, then it will necessarily be safe to delete CI/CD variable:
        1. If CI/CD variable doesn't exist: do nothing
        2. If CI/CD variable exists:
            1. CI/CD variables values doesn't match: notification of this (warning)
            2. CI/CD variables value matches: delete this CI/CD variable, notification of this (info)

    If <is_force> parameter is True, then it will necessarily be force to delete CI/CD variable:
        1. If CI/CD variable doesn't exist: do nothing
        2. If CI/CD variable exists:
            1. CI/CD variables values doesn't match: delete this file, notification of this (warning)
            2. CI/CD variables values matches: delete this file, notification of this (info)

    :param parameters: parameters that were passed to the package api
    :param variable: CI/CD variable to be deleted
    :param is_force: forced deletion (ignore variable value)
    """
    system_params = parameters._Parameters__system_params
    if variable['name'] in system_params.created_cicd_vars_names:
        raise PackageCICDVariablePolicyError(f"{variable['name']} CI/CD variable used in another package. According "
                                             f"to package manager policy, you cannot use the same variables in "
                                             f"different packages")

    is_force = True if is_force or system_params.is_force else False
    __delete_variable(
        system_params.project, variable, is_force=is_force
    )
    __delete_info_about_variable(variable)


def __add_info_about_variable(variable: CICDVariable) -> None:
    with open(__PACKAGE_API_RESULT, 'r') as file:
        content = json.load(file)
    if variable not in content['cicd_variables']:
        content['cicd_variables'].append(variable)
    with open(__PACKAGE_API_RESULT, 'w') as file:
        json.dump(content, file)


def __delete_info_about_variable(variable: CICDVariable) -> None:
    with open(__PACKAGE_API_RESULT, 'r') as file:
        content = json.load(file)
    if variable not in content['cicd_variables']:
        return
    index = content['cicd_variables'].index(variable)
    content['cicd_variables'].pop(index)
    with open(__PACKAGE_API_RESULT, 'w') as file:
        json.dump(content, file)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for local package download """
from typing import Iterable
from pathlib import Path
from pprint import pformat
from abc import ABC, abstractmethod
import tarfile
import json
import shutil
import re

import oras.client
import oras.container
import oras.auth
import oras.defaults
import requests
from requests.auth import HTTPBasicAuth
import yaml

from git_system_follower.variables import IMAGE_PACKAGE_MAP, PACKAGES_PATH, PACKAGE_DIRNAME
from git_system_follower.logger import logger
from git_system_follower.errors import (
    RemoteRepositoryError, DownloadPackageError, UnknownRegistryError, PackageNotFoundError
)
from git_system_follower.typings.cli import (
    PackageCLI, PackageCLITypes,
    PackageCLIImage, PackageCLITarGz, PackageCLISource
)
from git_system_follower.typings.registry import RegistryTypes, RegistryInfo
from git_system_follower.plugins.cli.packages.default import TarGzPlugin
from git_system_follower.typings.package import PackageLocalData
from git_system_follower.package.package_info import (
    DESCRIPTION_FILENAME, get_package_info, check_dependency_depth, add_dependencies
)
from git_system_follower.utils.tmpdir import tempdir


__all__ = ['download']


class Registry(ABC, oras.client.OrasClient):
    """ Base class of any new Registry """

    @abstractmethod
    def download(self, target: str, outdir: Path, *, registry: RegistryInfo) -> Path | None:
        """ Download package function from OCI artifact or image

        Similar command "oras copy <target> --to-oci-layout <outdir>" with custom:
        downloading regular docker image (<target>) with package inside it and unpacking this package from layer to
        <outdir>/<package name>@<package version>.tar.gz, where package name and version are taking from package.yaml
        inside package(.tar.gz file) itself

        :param target: OCI artifact or image url without protocol
        :param outdir: directory where need to save package
        :param registry: registry information like credentials for auth, insecure mode, etc.
        :return: content of downloaded OCI artifact or image
        """
        pass

    @abstractmethod
    def _basic_auth(self, container: oras.container.Container, *, registry: RegistryInfo) -> str:
        """ Get basic auth token for getting manifest and downloading layers using Basic
        Auth (`docker pull` simulation with possibility of anonymous image downloading)

        :param container: Oras container with information about target
        :param registry: registry information like credentials for auth, insecure mode, etc.
        :return: anonymous token
        """
        pass

    @staticmethod
    def _check_anonymous_token_request(response: requests.Response, container: oras.container.Container) -> str:
        """ Verification of the received anonymous token """
        response.raise_for_status()
        token = response.json().get('token')
        if token is None:
            raise RemoteRepositoryError(f'Failed to get an anonymous token for {container}')
        return token

    def get_manifest_wrapper(self, container: oras.container.Container) -> dict:
        """ Wrapper for getting manifest

        :param container: Oras container with information about target
        :return: manifest
        """
        manifest = self.get_manifest(
            container, allowed_media_type=['application/vnd.docker.distribution.manifest.v2+json']
        )
        logger.debug(f'Found manifest:\n{pformat(manifest)}')
        return manifest

    def is_gear(self, manifest: dict, container: oras.container.Container) -> bool:
        """ Check if the image is a GSF package

        :param manifest: manifest of OCI artifact or image
        :param container: Oras container with information about target
        """
        # OCI artifact
        if manifest['mediaType'] == 'application/vnd.oci.image.manifest.v1+json':
            return True

        # Docker image with: LABEL gsf.package="true"
        digest = manifest['config']['digest']
        response = requests.get(
            f'{self.prefix}://{container.registry}/v2/{container.namespace}/{container.repository}/blobs/{digest}',
            headers=self.auth.get_auth_header()
        )
        response.raise_for_status()
        labels = response.json().get('config', {}).get('Labels', {})
        if labels is None:  # case when label specified and is None
            return False

        required_label = 'gsf.package'
        return required_label in labels.keys() and labels[required_label] == 'true'

    def _download_layer(self, container: oras.container.Container, manifest: dict, outdir: Path) -> Path:
        """ Download layer of GSF package
        The GSF package has specific rules that there can only be one directory in an OCI artifact or image

        :param container: Oras container with information about target
        :param manifest: manifest of OCI artifact or image
        :param outdir: directory where need to save package
        :return: content of downloaded OCI artifact or image
        """
        layers = manifest.get("layers")
        if layers is None:
            raise DownloadPackageError(f'Failed to parse the layers of {container} image to get package '
                                       f'in it: layers not found')
        if len(layers) != 1:
            raise DownloadPackageError(f'{container} image should contain only one layer with the package. '
                                       f'The number of layers found: {len(layers)}')

        layer = layers[0]
        digest = layer['digest']
        package_path = outdir / 'package.tar.gz'
        self.download_blob(container, digest, str(package_path.absolute()))
        logger.debug(f'Downloaded {digest} layer to {package_path}')
        try:
            package_new_path = package_path.parent / self._get_package_filename(package_path)
        except DownloadPackageError as error:
            raise DownloadPackageError(f'{error}. Image: {container}, layer digest: {digest}')

        package_path.rename(package_new_path)
        logger.debug(f'Renamed {package_path} to {package_new_path}')
        return package_new_path

    @staticmethod
    def _get_package_filename(path: Path) -> str:
        """ Get current filename for package (.tar.gz file) using parsing pacakge.yaml inside this package

        :param path: path to package (.tar.gz)
        :return: current filename for package
        """
        name, version = get_name_and_version_version_from_targz(path)
        return f'{name}@{version}.tar.gz'


class RegistryV2(Registry):
    """ Oras client for downloading git-system-follower packages from docker registries
    using Docker Registry HTTP API V2 spec

    for more details, see https://docker-docs.uclv.cu/registry/spec/api/
    """

    def download(self, target: str, outdir: Path, *, registry: RegistryInfo) -> Path | None:
        if registry.is_insecure:
            self.prefix = "http"
        container = self.get_container(target)
        token = self._basic_auth(container, registry=registry)
        self.auth.set_token_auth(token)

        manifest = self.get_manifest_wrapper(container)
        if not self.is_gear(manifest, container):
            logger.warning(f'{target} is not a git-system-follower package. Skip downloading')
            return None

        return self._download_layer(container, manifest, outdir)

    def _basic_auth(self, container: oras.container.Container, *, registry: RegistryInfo) -> str:
        auth = None
        if registry.credentials is not None:
            auth = HTTPBasicAuth(registry.credentials.username, registry.credentials.password)
        response = requests.get(
            f'{self.prefix}://{container.registry}/v2/token',
            params={
                'service': container.registry,
                'scope': f'repository:{container.namespace}/{container.repository}:pull'
            }, auth=auth, verify=not registry.is_insecure
        )
        return self._check_anonymous_token_request(response, container)


class Dockerhub(RegistryV2):
    """ Oras client for downloading git-system-follower packages from DockerHub

    It also works as well as RegistryV2, but have custom logic
    (TODO: fix it using Docker Registry HTTP API V2 spec for a single solution for all registries)
    """

    def download(self, target: str, outdir: Path, *, registry: RegistryInfo) -> Path | None:
        container = self.get_container(target)
        token = self._basic_auth(container, registry=registry)
        self.auth.set_token_auth(token)

        # fix oras get_manifest for docker.io. For example, we set the image as docker.io/path/to/image:tag
        # then oras will try to download the image from www.docker.io, which is an invalid link.
        container.registry = 'index.docker.io'

        manifest = self.get_manifest_wrapper(container)
        if not self.is_gear(manifest, container):
            logger.warning(f'{target} is not a git-system-follower package. Skip downloading')
            return None

        return self._download_layer(container, manifest, outdir)

    def _basic_auth(self, container: oras.container.Container, *, registry: RegistryInfo) -> str:
        auth = None
        if registry.credentials is not None:
            auth = HTTPBasicAuth(registry.credentials.username, registry.credentials.password)
        response = requests.get(
            f'{self.prefix}://auth.{container.registry}/token',
            params={
                'service': f'registry.{container.registry}',
                'scope': f'repository:{container.namespace}/{container.repository}:pull'
            }, auth=auth, verify=not registry.is_insecure
        )
        return self._check_anonymous_token_request(response, container)


class Artifactory(RegistryV2):
    pass


class Nexus(RegistryV2):
    pass


class AwsEcr(RegistryV2):

    def download(self, target: str, outdir: Path, *, registry: RegistryInfo) -> Path | None:
        container = self.get_container(target)
        if registry.credentials is None:
            raise PermissionError('AWS ECR does not work in anonymous mode, credentials are required')
        self.auth = oras.auth.get_auth_backend('basic', self.session, registry.is_insecure)
        self.auth.set_basic_auth(registry.credentials.username, registry.credentials.password)

        manifest = self.get_manifest_wrapper(container)
        if not self.is_gear(manifest, container):
            logger.warning(f'{target} is not a git-system-follower package. Skip downloading')
            return None

        return self._download_layer(container, manifest, outdir)


def get_name_and_version_version_from_targz(path: Path) -> tuple[str, str]:
    """ Get name and version of package for .tar.gz archive

    :param path: path to package (.tar.gz)
    :return: current name and version of package
    """
    description = f'{PACKAGE_DIRNAME}/{DESCRIPTION_FILENAME}'
    with tarfile.open(path, 'r:gz') as tar:
        if description not in tar.getnames():
            raise DownloadPackageError(f'Could not find {description} file inside image')
        with tar.extractfile(description) as file:
            content = yaml.safe_load(file)

    return _get_name_and_version_from_description(content, description)


def get_name_and_version_version_from_source(path: Path) -> tuple[str, str]:
    """ Get name and version of package for source

    :param path: path to package (directory)
    :return: current name and version of package
    """
    description = f'{PACKAGE_DIRNAME}/{DESCRIPTION_FILENAME}'
    with open(path / description, 'r') as file:
        content = yaml.safe_load(file)

    return _get_name_and_version_from_description(content, description)


def _get_name_and_version_from_description(content: dict, description: str) -> tuple[str, str]:
    """ Get name and version of package from description package file

    :param content: content of description package file
    :return: current name and version of package
    """
    name, version = content.get('name'), content.get('version')
    if name is None:
        raise DownloadPackageError(f"Section 'name' not found in {description} file")
    if version is None:
        raise DownloadPackageError(f"Section 'version' not found in {description} file")
    return name, version


def download(
        packages: Iterable[PackageCLI | PackageCLIImage | PackageCLITarGz | PackageCLISource],
        directory: Path = PACKAGES_PATH, *,
        registry: RegistryInfo, dependency_tree: str = '', dependency_level: int = 0, is_deps_first: bool
) -> list[PackageLocalData]:
    """ Download packages

    :param packages: packages to be downloaded
    :param directory: directory where need to download package
    :param registry: registry information like credentials for auth, insecure mode, etc.
    :param dependency_tree: current dependency tree, e.g. `root-package -> root's-dependency`
    :param dependency_level: current dependency depth level
    :param is_deps_first: whether dependencies should be specified first, and then the main package. This is necessary
                          for the order in which package are handled, e.g. installation starts with dependencies,
                          uninstallation with the main package
    :return: paths to packages (.tar.gz files)
    """
    if not packages:
        return []

    if dependency_level == 0:
        logger.info(':: Downloading packages')

    result = []
    for i, package in enumerate(packages, 1):
        logger.info(f'-> Downloading {package}')
        new_dep_tree = f'{dependency_tree} -> {package.name}' if dependency_level != 0 else package.name
        check_dependency_depth(dependency_level, new_dep_tree)

        source = get_source(package, directory, registry=registry)
        if source is None:
            continue
        data = get_package_info(source.parent, source.name)
        if data['dependencies']:
            logger.info(f"Package dependencies: {', '.join([str(dep) for dep in data['dependencies']])}")
        dependencies_data = download(
            data['dependencies'], directory, registry=registry,
            dependency_tree=new_dep_tree, dependency_level=dependency_level + 1, is_deps_first=is_deps_first
        )
        fixed_dependency_names = []
        for dependency in data['dependencies']:
            fixed_dependency_names.append(_get_fixed_package_using_mapping(dependency))
        data['dependencies'] = tuple(fixed_dependency_names)
        result.append(data)
        result = add_dependencies(result, dependencies_data, is_deps_first)

    if dependency_level == 0:
        logger.success('Download complete')
    return result


def get_source(
        package: PackageCLI | PackageCLIImage | PackageCLITarGz | PackageCLISource, directory: Path, *,
        registry: RegistryInfo
) -> Path | None:
    """ Wrapper to handle different package input values

    :param package: packages to be downloaded
    :param directory: directory where need to download package
    :param registry: registry information like credentials for auth, insecure mode, etc.
    :return: source code (download package with unpacking or unpacking .tar.gz archive or already ready-made code)
    """
    if package.type == PackageCLITypes.source:
        source = package.path
        name, version = get_name_and_version_version_from_source(source)
        logger.info(f'{name}@{version} package is provided as source code (Path: {source})')
        if not source.exists():
            raise PackageNotFoundError(f'{source.absolute()} directory does not exist')
        return source / PACKAGE_DIRNAME

    if package.type == PackageCLITypes.image:
        path = download_package(package, directory, registry=registry)
        if path is None:
            return None
    else:  # package.type == PackageCLITypes.targz
        path = package.path
        name, version = get_name_and_version_version_from_targz(path)
        logger.info(f'{name}@{version} package is provided as .tar.gz archive (Path: {path})')
        if not path.exists():
            raise PackageNotFoundError(f'{path.absolute()} archive does not exist')

    return unpack(path, PACKAGES_PATH)


@tempdir
def download_package(
        package: PackageCLIImage, outdir: Path, *,
        tmpdir: Path, registry: RegistryInfo
) -> Path | None:
    """ Download package from registry using oras

    An explanation of how version detection works: we have the version inside package.yaml and the version (image tag)
    We take as true the version that is specified in package.yaml.

    That is, we definitely need to download the package to find out the correct version.
    To avoid repeated downloads we use a `IMAGE_PACKAGE_MAP` file where we specify which image
    corresponds to which package

    How IMAGE_PACKAGE_MAP file works:
        1. if package doesn't exist in `outdir` and package with image doesn't exist in `IMAGE_PACKAGE_MAP`:
           download package from image, save info in `IMAGE_PACKAGE_MAP`
        2. if package exists in `outdir`, but package with image doesn't exist in `IMAGE_PACKAGE_MAP`
           (e.g. this package/.tar.gz file was brought in from outside):
           download package from image to a temporary directory and skip moving to outdir,
           save info in `IMAGE_PACKAGE_MAP`
        3. if package doesn't exist in `outdir`, but package with image exists in `IMAGE_PACKAGE_MAP`:
           download package from image, save info in `IMAGE_PACKAGE_MAP`
        4. if package exists in `outdir` and package with image exists in `IMAGE_PACKAGE_MAP`:
           skip downloading package from image

    :param package: package to be downloaded
    :param outdir: directory where need to download package
    :param tmpdir: temporary directory where package will be downloaded and manipulated
    :param registry: registry information like credentials for auth, insecure mode, etc.
    :return: downloaded package (tar.gz file): `<outdir>/<package tar.gz file>`
    """
    if not IMAGE_PACKAGE_MAP.exists():
        with open(IMAGE_PACKAGE_MAP, 'w') as file:
            json.dump({}, file)

    outfile = _get_current_path_using_mapping(package, outdir)
    if outfile is not None:
        logger.info(f'{_get_filename_without_suffix(outfile)} package is provided as docker image (Image: {package})')
        logger.info(f'Package has already been downloaded to {outfile} from {package}. Skip downloading')
        return outfile

    client = get_client(package.registry, registry=registry)
    image = package.get_image_path()
    package_tmp_path = client.download(image, outdir=tmpdir, registry=registry)
    if package_tmp_path is None:  # image is not git-system-follower package
        return None

    outfile = outdir / package_tmp_path.name
    if outfile.exists():
        logger.warning(f'Package {outfile} already exist. Skip Moving {package_tmp_path} to {outfile}')
    else:
        shutil.move(package_tmp_path, outfile)
        logger.debug(f'Moved {package_tmp_path} to {outfile}')
    logger.info(f'{_get_filename_without_suffix(outfile)} package is provided as docker image (Image: {package})')
    logger.success(f'Downloaded package from {package} to {outfile}')
    _save_info_about_downloaded_package(package, outfile)
    return outfile


def _get_current_path_using_mapping(package: PackageCLI | PackageCLIImage, packages_dir: Path) -> Path | None:
    """ Get ratio of downloaded package to image from which they were obtained

    :param package: package to be downloaded
    :param packages_dir: directory with downloaded packages
    :return: path to downloaded package or None if no package is downloaded
    """
    with open(IMAGE_PACKAGE_MAP, 'r') as file:
        content: dict[str, str] = json.load(file)
    downloaded_packages = packages_dir.glob('*.tar.gz')
    for downloaded_package in downloaded_packages:
        filename = _get_filename_without_suffix(downloaded_package, suffix=TarGzPlugin.suffix)
        file_target = content.get(filename)
        if file_target is None:
            continue
        if file_target == str(package):
            return downloaded_package
    return None


def get_client(
        registry_address: str, *, registry: RegistryInfo
) -> Dockerhub | Artifactory | Nexus | AwsEcr:
    """ Identifies registry type and returns appropriate client

    :param registry_address: image registry (eg 'docker.io', 'artifactory.example.com:17001', 'nexus.host.com:16001')
    :param registry: registry information like credentials for auth, insecure mode, etc.
    :returns: Instance of Dockerhub, Artifactory, or Nexus client

    :raises UnknownRegistryError: registry could not be identified
    """
    scheme = 'http' if registry.is_insecure else 'https'
    if is_dockerhub(scheme, registry_address, registry_type=registry.type, is_insecure=registry.is_insecure):
        logger.info(f'{registry_address} is of type DockerHub')
        return Dockerhub(hostname=registry_address)

    if is_artifactory(scheme, registry_address, registry_type=registry.type, is_insecure=registry.is_insecure):
        logger.info(f'{registry_address} is of type Artifactory')
        return Artifactory(hostname=registry_address)

    if is_nexus(scheme, registry_address, registry_type=registry.type, is_insecure=registry.is_insecure):
        logger.info(f'{registry_address} is of type Nexus')
        return Nexus(hostname=registry_address)

    if is_awsecr(scheme, registry_address, registry_type=registry.type, is_insecure=registry.is_insecure):
        logger.info(f'{registry_address} is of type AWS ECR')
        return AwsEcr(hostname=registry_address)

    raise UnknownRegistryError(
        f'Could not determine the registry type for {registry_address}. '
        f'Supported registries: DockerHub, Artifactory, Nexus, AWS ECR'
    )


def is_dockerhub(scheme: str, registry: str, *, registry_type: RegistryTypes, is_insecure: bool) -> bool:
    """ Check if the given registry is Dockerhub """
    if registry_type == RegistryTypes.dockerhub:
        return True
    if registry_type != RegistryTypes.auto:
        return False

    url = f'{scheme}://index.{registry}/v2'
    try:
        response = requests.get(url, timeout=3, verify=not is_insecure)
    except requests.exceptions.RequestException as error:
        logger.debug(f'DockerHub REST API call error: {error}')
        return False
    # dockerhub always returns 401 code and 'www-authenticate' header in this end point
    return response.status_code == 401 and 'docker.io/token' in response.headers.get('www-authenticate', '')


def is_artifactory(scheme: str, registry: str, *, registry_type: RegistryTypes, is_insecure: bool) -> bool:
    """ Check if the given registry is Artifactory """
    if registry_type == RegistryTypes.artifactory:
        return True
    if registry_type != RegistryTypes.auto:
        return False

    url = f'{scheme}://{registry}/artifactory/api/system/version'
    try:
        response = requests.get(url, timeout=3, verify=not is_insecure)
    except requests.exceptions.RequestException as error:
        logger.debug(f'Artifactory REST API call error: {error}')
        return False
    return response.status_code == 200


def is_nexus(scheme: str, registry: str, *, registry_type: RegistryTypes, is_insecure: bool) -> bool:
    """ Check if the given registry is Nexus """
    if registry_type == RegistryTypes.nexus:
        return True
    if registry_type != RegistryTypes.auto:
        return False

    url = f'{scheme}://{registry}/v2/_catalog'
    try:
        response = requests.get(url, timeout=3, verify=not is_insecure)
    except requests.exceptions.RequestException as error:
        logger.debug(f'Nexus REST API call error: {error}')
        return False
    return response.status_code == 200 and 'Nexus' in response.headers.get('Server', '')


def is_awsecr(scheme: str, registry: str, *, registry_type: RegistryTypes, is_insecure: bool) -> bool:
    """ Check if the given registry is Aws ECR """
    if registry_type == RegistryTypes.awsecr:
        return True
    if registry_type != RegistryTypes.auto:
        return False

    url = f'{scheme}://{registry}'
    try:
        response = requests.get(url, timeout=3, verify=not is_insecure)
    except requests.exceptions.RequestException as error:
        logger.debug(f'AWS API call error: {error}')
        return False

    regex = re.compile(r'(\w+)=["\']([^"\']+)["\']')
    auth_header = response.headers.get('Www-Authenticate', '')
    headers = dict(regex.findall(auth_header))
    return headers.get('service', '') == 'ecr.amazonaws.com'


def _save_info_about_downloaded_package(package: PackageCLI, current_package: Path) -> None:
    """ Save name and version mapping information from package.yaml with docker image name and tag

    :param package: package to be downloaded (docker image)
    :param current_package: downloaded package (tar.gz file)
    """
    with open(IMAGE_PACKAGE_MAP, 'r') as file:
        content: dict[str, str] = json.load(file)
    current_package_filename = _get_filename_without_suffix(current_package, suffix=TarGzPlugin.suffix)
    content[current_package_filename] = str(package)
    with open(IMAGE_PACKAGE_MAP, 'w') as file:
        json.dump(content, file, indent=4)


def _get_filename_without_suffix(path: Path, suffix: str = TarGzPlugin.suffix) -> str:
    """ Get file name without `suffix`

    :param path: file path
    :param suffix: file suffix, e.g. `.tar.gz`
    :return: file name without `suffix`
    """
    return path.name[:-len(suffix)]


def unpack(path: Path, outdir: Path) -> Path:
    """ Unpack package/.tar.gz file

    :param path: package/.tar.gz file path
    :param outdir: output directory
    :return: sources path of this package
    """
    filename = _get_filename_without_suffix(path, suffix=TarGzPlugin.suffix)
    package_dir = outdir / filename / PACKAGE_DIRNAME
    if package_dir.exists():
        return package_dir

    with tarfile.open(path) as tar:
        tar.extractall(package_dir.parent)
    return package_dir


def _get_fixed_package_using_mapping(package: PackageCLIImage | PackageCLITarGz | PackageCLISource) -> PackageCLI:
    """ Get ratio of downloaded package to image from which they were obtained

    :param package: package to be downloaded
    :return: path to downloaded package or None if no package is downloaded
    """
    with open(IMAGE_PACKAGE_MAP, 'r') as file:
        content: dict[str, str] = json.load(file)
    for key, value in content.items():
        if value == str(package):
            name, version = key.split('@')
            return PackageCLI(name=name, version=version)
    raise DownloadPackageError(f'Failed to parse dependency package name, version for {package}')



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from dataclasses import dataclass
from typing import Optional

from git_system_follower.typings.cli import Credentials


__all__ = ['RegistryTypes', 'RegistryInfo']


class RegistryTypes(Enum):
    auto = 'Autodetect'
    dockerhub = 'Dockerhub'
    nexus = 'Nexus'
    artifactory = 'Artifactory'
    awsecr = 'AwsEcr'


@dataclass
class RegistryInfo:
    credentials: Optional[Credentials]
    type: RegistryTypes
    is_insecure: bool



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from git import Repo

import gitlab

from git_system_follower.logger import logger
from git_system_follower.typings.repository import RepositoryInfo


__all__ = ['checkout_to_new_branch', 'push_installed_packages']


def checkout_to_new_branch(repo: Repo, base_branch: str) -> str:
    new_branch = f'{base_branch}.temp-manage-packages'
    repo.git.checkout(base_branch)
    repo.remotes.origin.pull(base_branch)

    for branch in repo.heads:
        if branch.name == new_branch:
            repo.delete_head(new_branch, force=True)
            logger.debug(f'Local branch {new_branch} deleted')
            break

    repo.git.checkout('-b', new_branch)
    logger.success(f'Created new {new_branch} local branch (local repo: {repo.git.working_dir})')
    return new_branch


def push_installed_packages(repo: RepositoryInfo, msg: str, *, name: str, email: str) -> None:
    """ Push changes to remote repository

    :param repo: repo information
    :param msg: commit message
    :param name: user name for commit changes
    :param email: user email for commit changes
    """
    try:
        repo.gitlab.branches.delete(repo.git.active_branch.name)
    except gitlab.exceptions.GitlabDeleteError:
        # if temp branch does not exist
        pass

    repo.git.git.add(A=True)

    repo.git.config_writer().set_value('user', 'name', name).release()
    repo.git.config_writer().set_value('user', 'email', email).release()
    repo.git.index.commit(msg)

    repo.git.remotes.origin.push(repo.git.active_branch.name)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import os


ROOT_DIR = Path('.git-system-follower').absolute()
REPOS_PATH = ROOT_DIR / 'repositories'
PACKAGES_PATH = ROOT_DIR / 'packages'
if not PACKAGES_PATH.exists():
    os.makedirs(PACKAGES_PATH)

PACKAGE_API_RESULT = ROOT_DIR / 'package-api-result.json'
IMAGE_PACKAGE_MAP = ROOT_DIR / 'image-package-map.json'

PACKAGE_DIRNAME = 'git-system-follower-package'
SCRIPTS_DIR = 'scripts'
PACKAGE_DESCRIPTION_FILE_API = {
    'v1': {
        'mandatory_sections': ('apiVersion', 'type', 'name', 'version'),
        'optional_sections': ('dependencies',),
        'section_types': (str, str, str, str, {list: str}),
        'package_types': ('gitlab-ci-pipeline',)
    }
}



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from gitlab.v4.objects import Project

from git_system_follower.logger import logger
from git_system_follower.variables import PACKAGE_DIRNAME, SCRIPTS_DIR
from git_system_follower.typings.package import PackageLocalData
from git_system_follower.typings.repository import RepositoryInfo
from git_system_follower.states import PackageState
from git_system_follower.typings.cli import ExtraParam
from git_system_follower.typings.script import ScriptResponse
from git_system_follower.package.script import run_script
from git_system_follower.package.cicd_variables import CICDVariable, get_cicd_variables


__all__ = ['delete']


def delete(
        package: PackageLocalData, repo: RepositoryInfo, state: PackageState, *,
        created_cicd_variables: tuple[str, ...], extras: tuple[ExtraParam, ...], is_force: bool
) -> ScriptResponse:
    logger.info('==> Package deletion')
    workdir = Path(repo.git.working_dir)
    scripts_dir = package['path'] / PACKAGE_DIRNAME / SCRIPTS_DIR / package['version']
    if not scripts_dir.exists():
        raise FileNotFoundError(f'Scripts directory is missing ({scripts_dir.absolute()})')

    current_cicd_variables = get_cicd_variables(repo.gitlab)
    response = run_delete_script(
        scripts_dir, workdir, repo.gitlab, current_cicd_variables, state,
        created_cicd_variables=created_cicd_variables, extras=extras, is_force=is_force
    )
    logger.success(f"Uninstalled {package['name']}@{package['version']} package")
    return response


def run_delete_script(
        script_dir: Path, workdir: Path, project: Project, current_cicd_variables: dict[str, CICDVariable],
        state: PackageState, *,
        created_cicd_variables: tuple[str, ...], extras: tuple[ExtraParam, ...], is_force: bool
) -> ScriptResponse:
    logger.info('\tRunning delete package api')
    path = script_dir / 'delete.py'
    response = run_script(
        path, workdir, project, current_cicd_variables, state['used_template'],
        created_cicd_variables=created_cicd_variables, extras=extras, is_force=is_force, state=state
    )
    return response



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from packaging.version import Version


__all__ = ['normalize_version']


def normalize_version(version: str) -> Version:
    # Remove prefixes and replace characters
    # Examples:
    #   v24.4_r1ю0ю0 -> 24.4.1.0.0
    #   v2alpha1 -> 2.1
    #   v3-beta-4 -> 3.4
    #   1.0.0 -> 1.0.0
    version = re.sub(r'[a-zA-Z]+', '.', version.lstrip('v'))
    version = re.sub(r'[._-]+', '.', version)
    return Version(version.strip('.'))



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from urllib.parse import urlparse, urlunparse

from git import Repo
from git.config import GitConfigParser
from gitlab.v4.objects import Project

from git_system_follower.logger import logger
from git_system_follower.variables import REPOS_PATH
from git_system_follower.typings.cli import PackageCLI
from git_system_follower.typings.package import PackageLocalData


__all__ = ['get_config', 'get_packages_str', 'get_git_repo']


def get_config(path: str):
    path = Path(path).expanduser()
    return GitConfigParser(path, read_only=True)


def get_packages_str(packages: tuple[PackageLocalData | PackageCLI, ...]) -> str:
    if isinstance(packages[0], dict):
        return ', '.join([f"{package['name']}@{package['version']}" for package in packages])
    return ', '.join([str(package) for package in packages])


def get_git_repo(project: Project, token: str) -> Repo:
    """ Clone/get git repo

    :param project: gitlab project
    :param token: access token
    """
    url = project.http_url_to_repo
    directory = get_repo_directory_path(url)
    if directory.exists():
        logger.debug(f'Local repository {directory} is used')
        repo = Repo(directory)
        repo.remotes.origin.fetch()
        return repo
    url = get_url_with_token(url, token)
    repo = Repo.clone_from(url, directory)
    logger.debug(f'Repository cloned into {directory} folder')
    return repo


def get_repo_directory_path(repo_url: str) -> Path:
    name = repo_url.split('/')[-1].replace('.git', '')
    return REPOS_PATH / name


def get_url_with_token(url: str, token: str) -> str:
    url = urlparse(url)
    url = url._replace(netloc=f'auth2:{token}@{url.netloc}')
    return str(urlunparse(url))



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for working with state file """
from typing import TypedDict, Any, NamedTuple
from enum import Enum
from pathlib import Path
import hashlib
from datetime import datetime
from pprint import pformat
import base64

import yaml

from git_system_follower.logger import logger
from git_system_follower.errors import HashesMismatch
from git_system_follower.typings.cli import PackageCLI
from git_system_follower.typings.package import PackageLocalData
from git_system_follower.typings.script import ScriptResponse
from git_system_follower.package.cicd_variables import CICDVariable


__all__ = [
    'ChangeStatus', 'PackageState', 'StateFile',
    'get_installed_packages', 'filter_cicd_variables_by_state', 'update_created_cicd_variables',
    'mask_data', 'unmask_data'
]


class ChangeStatus(Enum):
    no_change = 0
    changed = 1


class CICDVariablesSection(TypedDict):
    names: list[str]
    hash: str


class PackageState(TypedDict):
    name: str
    version: str
    used_template: str
    template_variables: dict[str, str]
    last_update: str
    dependencies: list[str]
    cicd_variables: CICDVariablesSection


class StateFileContent(TypedDict):
    hash: str
    packages: list[PackageState]


class InstalledPackage(NamedTuple):
    name: str
    version: str


class StateFile:
    __name = '.state.yaml'

    def __init__(self, *, raw: bytes | None = None, current_cicd_variables: dict[str, CICDVariable] | None = None):
        """ Read raw state file (e.g. from GitLab REST API) or init state file with empty packages section

        :param raw: state file
        :param current_cicd_variables: current CI/CD variables in Gitlab
        """
        self.__change_status = ChangeStatus.no_change
        if raw is None:
            packages = []
            computed_hash = self.__get_hash(packages)
            self.__content = StateFileContent(hash=computed_hash, packages=packages)
            return

        content: StateFileContent = yaml.safe_load(raw)
        computed_hash = self.__get_hash(content['packages'])
        if content['hash'] != computed_hash:
            raise HashesMismatch(f"Hash specified in state file ({content['hash']}) and "
                                 f"generated hash ({computed_hash}) do not match",
                                 state_file_hash=content['hash'], generated_hash=computed_hash)

        for package in content['packages']:
            self.__check_cicd_variables_hash(package, current_cicd_variables)
        self.__content = StateFileContent(hash=computed_hash, packages=content['packages'])

    def __get_hash(self, state: Any) -> str:
        """ Generate hash for any variable.
        For example, for 'packages' section in state file, for 'cicd_variables' section

        :param state: state ('packages' section) from state file

        :return: generated hash
        """
        sorted_state = self.__sort_state(state)  # sort for the same behaviour when working with hash for saving/reading
        string = str(sorted_state)
        return hashlib.sha256(string.encode()).hexdigest()

    def __sort_state(self, state: list[PackageState] | PackageState):
        if isinstance(state, list):
            return [self.__sort_state(item) for item in state]
        elif isinstance(state, dict):
            return {key: self.__sort_state(value) for key, value in sorted(state.items())}
        else:
            return state

    def __check_cicd_variables_hash(
            self, package: PackageState, current_cicd_variables: dict[str, CICDVariable]
    ) -> None:
        """ Check hash for CI/CD variables of <package>

        :param package: package with information about variable names
        :param current_cicd_variables: current CI/CD variables in Gitlab
        """
        variables = filter_cicd_variables_by_state(package, current_cicd_variables)
        computed_hash = self.__get_hash(variables)
        if computed_hash != package['cicd_variables']['hash']:
            raise HashesMismatch(f"CI/CD variables hash specified in state file in "
                                 f"{package['name']}@{package['version']} package "
                                 f"({package['cicd_variables']['hash']}) "
                                 f"and generated hash ({computed_hash}) do not match",
                                 state_file_hash=package['cicd_variables']['hash'], generated_hash=computed_hash)

    def get_installed_packages(self) -> tuple[InstalledPackage, ...]:
        packages = []
        for package in self.__content['packages']:
            packages.append(InstalledPackage(name=package['name'], version=package['version']))
        return tuple(packages)

    def get_all_created_cicd_variables(self) -> tuple[str, ...]:
        """ Get created CI/CD variables from state file from all packages

        :return: list of CI/CD variables names in all installed packages
        """
        variables = []
        for package in self.__content['packages']:
            variables.extend(package['cicd_variables']['names'])
        return tuple(variables)

    def get_package(self, package: PackageLocalData, *, for_delete: bool) -> PackageState | None:
        """ Get state with package from state

        :param package: package which need to find in states
        :param for_delete: is need to find package to delete (or to install)

        :return: found package
        """
        if for_delete:
            return self.__get_package_state_by_name_and_version(package)
        return self.__get_package_state_by_name(package)

    def __get_package_state_by_name_and_version(self, package: PackageLocalData) -> PackageState | None:
        """ Get state with package from state by name

        :param package: package which need to find in states

        :return: found package by name and version
        """
        for state in self.__content['packages']:
            if package['name'] == state['name'] and package['version'] == state['version']:
                return state

    def __get_package_state_by_name(self, package: PackageLocalData) -> PackageState | None:
        """ Get state with package from state by name

        :param package: package which need to find in states

        :return: found package by name
        """
        for state in self.__content['packages']:
            if package['name'] == state['name']:
                return state

    def add_package(
            self, package: PackageLocalData, response: ScriptResponse | None, state: PackageState | None
    ) -> None:
        """ Add package to state file

        :param package: package which need to add to state file
        :param response: script response with information about used template, used ci/cd variables
        :param state: current state from state file (if package already installed but another versions)
        """
        if response is None:
            return

        self.__change_status = ChangeStatus.changed
        variables_names = [variable['name'] for variable in response['cicd_variables']]
        new_state = PackageState(
            name=package['name'], version=package['version'],
            used_template=response['template'],
            template_variables={name: mask_data(value) for name, value in response['template_variables'].items()},
            last_update=str(datetime.now()),
            dependencies=[f"{dependency.name}@{dependency.version}" for dependency in package['dependencies']],
            cicd_variables=CICDVariablesSection(
                names=variables_names,
                hash=self.__get_hash(response['cicd_variables'])
            )
        )
        if state is None:
            self.__content['packages'].append(new_state)
            return
        index = self.__content['packages'].index(state)
        self.__content['packages'][index] = new_state

    def delete_package(self, state: PackageState) -> PackageState:
        self.__change_status = ChangeStatus.changed
        index = self.__content['packages'].index(state)
        return self.__content['packages'].pop(index)

    def get_packages(self) -> list[PackageState]:
        return self.__content['packages'].copy()

    def status(self) -> ChangeStatus:
        return self.__change_status

    def save(self, directory: Path) -> None:
        """ Save state file

        :param directory: path where state file will be saved
        """
        path = directory / self.__name
        state = self.__content['packages']
        computed_hash = self.__get_hash(state)
        content = StateFileContent(hash=computed_hash, packages=state)
        logger.debug(f'New hash generated: {computed_hash}')
        with open(path, 'w') as file:
            yaml.dump(content, file)

    def __str__(self) -> str:
        return pformat(self.__content['packages'])


def filter_cicd_variables_by_state(
        state: PackageState | None, current_cicd_variables: dict[str, CICDVariable]
) -> list[CICDVariable]:
    """ Get current CI/CD variables of package state for CI/CD variables specified in state file
    Note: only variable names are stored in state file

    :param state: state from state file with variable names for which necessary to find it's current state
    :param current_cicd_variables: current CI/CD variables in Gitlab
    :return: list of CI/CD variables filtered by necessary variable names
    """
    if state is None:
        return []

    names = state['cicd_variables']['names']
    variables = []
    for variable in names:
        if variable in current_cicd_variables.keys():
            variables.append(current_cicd_variables[variable])
    return variables


def get_installed_packages(states: dict[str, StateFile]) -> set[PackageCLI]:
    """ Getting information about installed packages

    :param states: current states in GitLab repository branches
    :return: installed packages set
    """
    installed_packages = set()
    for package_states in states.values():
        for package in package_states.get_installed_packages():
            installed_packages.add(PackageCLI(name=package.name, version=package.version))
    return installed_packages


def update_created_cicd_variables(
        created_cicd_variables: list[str], response: ScriptResponse
) -> list[str]:
    response_variables = response['cicd_variables'] if response is not None else []
    for variable in response_variables:
        created_cicd_variables.append(variable['name'])
    return created_cicd_variables


def mask_data(data: str) -> str:
    encoded_bytes = base64.b64encode(data.encode("utf-8"))
    return encoded_bytes.decode("utf-8")


def unmask_data(encoded_data: str) -> str:
    decoded_bytes = base64.b64decode(encoded_data)
    return decoded_bytes.decode("utf-8")



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with defaults for package api: init.py, delete.py """

from git_system_follower.develop.api.types import Parameters
from git_system_follower.develop.api.templates import get_template_names, create_template, delete_template


def init_default_main(parameters: Parameters):
    templates = get_template_names(parameters)
    if not templates:
        raise ValueError('There are no templates in the package')

    if len(templates) > 1:
        template = parameters.extras.get('TEMPLATE')
        if template is None:
            raise ValueError('There are more than 1 template in the package, '
                             'specify which one you want to use with the TEMPLATE variable')
    else:
        template = templates[0]

    variables = parameters.extras.copy()
    variables.pop('TEMPLATE', None)
    create_template(parameters, template, variables)


def delete_default_main(parameters: Parameters):
    delete_template(parameters)




# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with plugin managers

It is necessary to load all managers at the start of the program to catch any plugin-related errors immediately.
Therefore, they are declared right away rather than when they are needed
"""

from git_system_follower.plugins.cli.packages.manager import PluginManager as CliPackagesPluginManager


__all__ = ['managers', 'cli_packages_pm']


cli_packages_pm = CliPackagesPluginManager()

managers = [cli_packages_pm]



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api to work with templates for package developers """
import json
from typing import Optional

from git_system_follower.develop.api.types import Parameters, ExtraParams, ExtraParam
from git_system_follower.variables import PACKAGE_API_RESULT as __PACKAGE_API_RESULT
from git_system_follower.errors import PackageAPIDevelopmentError, PackageTemplatePolicyError
from git_system_follower.package.templates import (
    get_template_names as __get_template_names,
    create_template as __create_template,
    delete_template as __delete_template
)


__all__ = ['get_template_names', 'create_template', 'update_template', 'delete_template']


def get_template_names(parameters: Parameters) -> tuple[str, ...]:
    """ Get available template names

    :param parameters: parameters that were passed to the package api
    :returns: tuple of template names
    """
    system_params = parameters._Parameters__system_params
    return __get_template_names(system_params.script_dir)


def create_template(
        parameters: Parameters, name: str, variables: Optional[ExtraParams | dict[str, str]] = None, *,
        is_force: bool = False
) -> None:
    """ Create files using cookiecutter template

    If <is_force> parameter is False, then it will necessarily be safe to create files using template:
        1. If file doesn't exist: create file
        2. If file exists:
            1. Files content doesn't match: notification of this (warning)
            2. Files content matches: notification of this (info)

    If <is_force> parameter is True, then it will necessarily be force to create files using template:
        1. If file doesn't exist: create file
        2. If file exists:
            1. Files content doesn't match: overwrite this file, notification of this (warning)
            2. Files content matches: notification of this (info)

    :param parameters: parameters that were passed to the package api
    :param name: name of template to be created
    :param variables: A dict of parameters used to generate the template. Which will be saved in the state file
    :param is_force: forced creation (ignore file content)
    """
    system_params = parameters._Parameters__system_params
    is_force = True if is_force or system_params.is_force else False
    variables = __get_variables(variables)
    __add_info_about_template(name, variables)
    __create_template(
        system_params.script_dir, name, parameters.workdir,
        variables=variables, is_force=is_force
    )


def update_template(
        parameters: Parameters, variables: Optional[ExtraParams | dict[str, str]] = None, *,
        is_force: bool = False
) -> None:
    """ Update files using cookiecutter template

    If <is_force> parameter is False, then it will necessarily be safe to update files using template:
        1. If file doesn't exist: create file
        2. If file exists: do nothing
            1. Files content doesn't match: notification of this (warning)
            2. Files content matches: notification of this (info)

    If <is_force> parameter is True, then it will necessarily be force to update files using template:
        1. If file doesn't exist: create file
        2. If file exists:
            1. Files content doesn't match: overwrite this file, notification of this (warning)
            2. Files content matches: notification of this (info)

    :param parameters: parameters that were passed to the package api
    :param variables: A dict of parameters used to generate the template. Which will be saved in the state file
    :param is_force: forced creation (ignore file content)
    """
    system_params = parameters._Parameters__system_params
    is_force = True if is_force or system_params.is_force else False
    variables = __update_template_variables(variables)
    __create_template(
        system_params.script_dir, parameters.used_template, parameters.workdir,
        variables=variables, is_force=is_force
    )


def delete_template(parameters: Parameters, *, is_force: bool = False) -> None:
    """ Delete files using cookiecutter template

    If <is_force> parameter is False, then it will necessarily be safe to delete files using template:
        1. If file doesn't exist: do nothing
        2. If file exists:
            1. Files content doesn't match: notification of this (warning)
            2. Files content matches: delete this file, notification of this (info)

    If <is_force> parameter is True, then it will necessarily be force to delete files using template:
        1. If file doesn't exist: do nothing
        2. If file exists:
            1. Files content doesn't match: delete this file, notification of this (warning)
            2. Files content matches: delete this file, notification of this (info)

    :param parameters: parameters that were passed to the package api
    :param is_force: forced deletion (ignore file content)
    """
    system_params = parameters._Parameters__system_params
    is_force = True if is_force or system_params.is_force else False
    __delete_info_about_template()
    __delete_template(
        system_params.script_dir, parameters.used_template, parameters.workdir,
        variables=parameters.template_variables, is_force=is_force
    )


def __add_info_about_template(template: str, variables: dict[str, str]) -> None:
    with open(__PACKAGE_API_RESULT, 'r') as file:
        content = json.load(file)
    if content['template'] is not None:
        raise PackageTemplatePolicyError('You cannot create multiple templates, only one')
    content['template'] = template
    content['template_variables'] = variables
    with open(__PACKAGE_API_RESULT, 'w') as file:
        json.dump(content, file)


def __update_template_variables(variables: dict[str, str]) -> dict[str, str]:
    with open(__PACKAGE_API_RESULT, 'r') as file:
        content = json.load(file)
    template_variables = content['template_variables'].copy()
    template_variables.update(__get_variables(variables))
    content['template_variables'] = template_variables
    with open(__PACKAGE_API_RESULT, 'w') as file:
        json.dump(content, file)
    return template_variables


def __delete_info_about_template() -> None:
    with open(__PACKAGE_API_RESULT, 'r') as file:
        content = json.load(file)
    if content['template'] is None:
        raise PackageTemplatePolicyError('No template was used, nothing to delete')
    content['template'] = None
    content['template_variables'] = []
    with open(__PACKAGE_API_RESULT, 'w') as file:
        json.dump(content, file)


def __get_variables(variables: Optional[ExtraParams | dict[str, str]] = None) -> dict[str, str]:
    """ Get variables for generate template and save these variables to state file

    Convert ExtraParams to dict[str, str]

    :param variables: A dict of parameters used to generate the template. Which will be saved in the state file
    :return: dict of variables
    """
    if variables is None:
        return {}

    if not isinstance(variables, dict):
        raise PackageAPIDevelopmentError(f"Invalid variables type. 'dict' or 'ExtraParams' is required. "
                                         f"Current type={type(variables)}")
    elif all(isinstance(k, str) and isinstance(v, str) for k, v in variables.items()):
        return variables
    elif all(isinstance(k, str) and isinstance(v, ExtraParam) for k, v in variables.items()):
        return {variable.name: variable.value for variable in variables.values()}
    else:
        raise PackageAPIDevelopmentError('Invalid variables value type')



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with cli types """
from dataclasses import dataclass
from typing import NamedTuple
from pathlib import Path
from enum import Enum


__all__ = [
    'PackageCLITypes', 'PackageCLI',
    'PackageCLIImage', 'PackageCLITarGz', 'PackageCLISource',
    'ExtraParam', 'Credentials'
]


# Types for packages
class PackageCLITypes(Enum):
    unknown = 0
    image = 1
    targz = 2
    source = 3


@dataclass(frozen=True, kw_only=True)
class PackageCLI:
    type: PackageCLITypes = PackageCLITypes.unknown
    name: str | None = None
    version: str | None = None

    def __str__(self):
        return f'{self.name}@{self.version}'


@dataclass(frozen=True, kw_only=True)
class PackageCLIImage(PackageCLI):
    type: PackageCLITypes = PackageCLITypes.image

    registry: str
    repository: str
    image: str
    tag: str = 'latest'

    def get_image_path(self) -> str:
        """ Get image path (without registry) """
        path = f'{self.repository}/' if self.repository else ''
        return f'{path}{self.image}:{self.tag}'

    def __str__(self):
        return f'{self.registry}/{self.get_image_path()}'


@dataclass(frozen=True, kw_only=True)
class PackageCLITarGz(PackageCLI):
    type: PackageCLITypes = PackageCLITypes.targz

    path: Path

    def __str__(self):
        return str(self.path)


@dataclass(frozen=True, kw_only=True)
class PackageCLISource(PackageCLI):
    type: PackageCLITypes = PackageCLITypes.source

    path: Path

    def __str__(self):
        return str(self.path)


class ExtraParam(NamedTuple):
    name: str
    value: str
    masked: bool


class Credentials(NamedTuple):
    username: str
    password: str
    # for adding token necessary to think use this class or create several class,
    # e.g. Token = str, Auth = Optional[Union[Credentials, Token]]



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for executing package api """
from pathlib import Path
from enum import Enum
import importlib.util
import os
import sys
import json

from gitlab.v4.objects import Project

from git_system_follower.errors import PackageApiError
from git_system_follower.variables import PACKAGE_API_RESULT
from git_system_follower.typings.cli import ExtraParam
from git_system_follower.package.cicd_variables import CICDVariable
from git_system_follower.states import PackageState, filter_cicd_variables_by_state, unmask_data
from git_system_follower.develop.api.types import Parameters, SystemParameters, ExtraParams
from git_system_follower.package.system import get_system_info
from git_system_follower.typings.script import ScriptResponse
from git_system_follower.package.default import init_default_main, delete_default_main


__all__ = ['run_script']


class SubprocessStatus(Enum):
    ok = 0


def run_script(
        path: Path, workdir: Path, project: Project, all_cicd_variables: dict[str, CICDVariable],
        used_template: str | None, *,
        extras: tuple[ExtraParam, ...], is_force: bool, state: PackageState | None = None,
        created_cicd_variables: tuple[str, ...]
) -> ScriptResponse:
    """ Run script (package api): init/update/delete

    :param path: path to package api
    :param workdir: workdir for package api
    :param project: gitlab project
    :param all_cicd_variables: all CI/CD variables in Gitlab repository
    :param used_template: last used template
    :param created_cicd_variables: list of created CI/CD variables in previous package installations
    :param extras: extra parameters to be passed to package api
    :param is_force: forced working: create/delete CI/CI variables, create/delete template, etc.
    :param state: current state for this package with another version
    :return: script response with information: CI/CD variables, last used template, etc. after running package api
    """
    template_variables = get_template_variables(state)
    cicd_variables = filter_cicd_variables_by_state(state, all_cicd_variables)
    created_cicd_vars_in_other_pkgs = _fetch_cicd_vars_except_package(created_cicd_variables, cicd_variables)
    default_package_api = init_default_main if path.name == 'init.py' else delete_default_main
    response = execute_package_api(
        path, workdir, project, used_template,
        template_variables=template_variables,
        cicd_variables={variable['name']: variable for variable in cicd_variables},
        all_cicd_variables=all_cicd_variables, created_cicd_vars_in_other_pkgs=created_cicd_vars_in_other_pkgs,
        extras=extras, is_force=is_force,
        default=default_package_api
    )
    return response


def get_template_variables(state: PackageState | None) -> dict[str, str]:
    if state is None:
        return {}

    return {name: unmask_data(value) for name, value in state['template_variables'].items()}


def _fetch_cicd_vars_except_package(all_vars_names: tuple[str, ...], pkg_variables: list[CICDVariable]) -> list[str]:
    """ Get created CI/CD variables except current package

    :param all_vars_names: all CI/CD variables names created in repository
    :param pkg_variables: CI/CD variables created in current package of another version
    :return: CI/CD variables names except package CI/CD variable
    """
    result = []
    pkg_vars_names = [package['name'] for package in pkg_variables]
    for variable in all_vars_names:
        if variable not in pkg_vars_names:
            result.append(variable)
    return result


def execute_module(func):
    """ Wrapper for executing package api: import module, change workdir """
    def wrapper(path: Path, workdir: Path, *args, default: str | None = None, **kwargs):
        """ Wrapper for execute_module decorator

        :param path: path to package api
        :param workdir: working directory for package api
        :param default: default package api code if package api doesn't exist (for init.py, delete.py)
        """
        path = path.absolute()
        workdir = workdir.absolute()

        module = _load_module(path, default=default)
        old = os.getcwd()
        os.chdir(workdir)
        result = func(path, workdir, *args, **kwargs, module=module)
        os.chdir(old)
        os.remove(PACKAGE_API_RESULT)
        return result
    return wrapper


def _load_module(path: Path, *, default: str | None):
    if not path.exists():
        if path.name == 'update.py':
            raise PackageApiError(f'No script file. Path: {path}')
        if default is None:
            raise PackageApiError(f'Module {path} not found and no default provided')
        module = type('default_module', (), {})
        module.main = default
        return module

    # add the path with the api package so that relative import can work
    module_dir = str(path.parent)
    if module_dir not in sys.path:
        sys.path.append(module_dir)

    # add module for running package api
    spec = importlib.util.spec_from_file_location('package_api', path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    func_name = 'main'
    if func_name not in dir(module):
        raise PackageApiError(f"The '{func_name}' function is missing in the module {path.absolute()}")
    if module.main.__class__.__name__ != 'function':
        raise PackageApiError(f"Object '{func_name}' is not a function in the module {path.absolute()}")
    return module


@execute_module
def execute_package_api(
        path: Path, workdir: Path, project: Project, used_template: str | None, *,
        module,
        template_variables: dict[str, str],
        cicd_variables: dict[str, CICDVariable], all_cicd_variables: dict[str, CICDVariable],
        created_cicd_vars_in_other_pkgs: list[str], extras: tuple[ExtraParam, ...], is_force: bool
) -> ScriptResponse:
    with open(PACKAGE_API_RESULT, 'w') as file:
        json.dump({
                'template': used_template,
                'template_variables': template_variables,
                'cicd_variables': [value for key, value in cicd_variables.items()]
            },
            file
        )

    system_params = SystemParameters(
        project=project, created_cicd_vars_names=created_cicd_vars_in_other_pkgs,
        script_dir=path.parent.absolute(), is_force=is_force
    )
    extras = get_remodeled_extras(extras)
    params = Parameters(
        _Parameters__system_params=system_params,  # use private class field
        system=get_system_info(project, extras),
        workdir=workdir.absolute(),
        extras=extras,
        cicd_variables=cicd_variables,
        all_cicd_variables=all_cicd_variables,
        used_template=used_template,
        template_variables=template_variables
    )
    module.main(params)

    with open(PACKAGE_API_RESULT, 'r') as file:
        content: ScriptResponse = json.load(file)
    return content


def get_remodeled_extras(extras: tuple[ExtraParam, ...]) -> ExtraParams:
    """ Get remodeled extra parameters for convenience of taking parameter by name

    from model: (
        ExtraParam(name='first_var', value='first_value', masked=True),
        ExtraParam(name='second_var', value='second_value', masked=False)
    )

    to model: {
        'first_var': ExtraParam(name='first_var', value='first_value', masked=True),
        'second_var': ExtraParam(name='second_var', value='second_value', masked=False)
    }

    :param extras: extra parameters to be passed to package api
    :return: remodeled extra parameters
    """
    return {extra.name: extra for extra in extras}



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import tempfile
import shutil

from git_system_follower.logger import logger


__all__ = ['tempdir']


def tempdir(func):
    """ Creating temporary directory while executing function """
    def wrapper(*args, **kwargs):
        directory = Path(tempfile.mkdtemp(prefix='gsf-package-manager-'))
        logger.debug(f'Temporary directory was created at {directory}')
        result = func(*args, **kwargs, tmpdir=directory)
        shutil.rmtree(directory)
        logger.debug(f'Temporary directory was removed at {directory}')
        return result
    return wrapper



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with output info related utils """
from typing import Mapping, Iterable, Callable, Sequence
import textwrap

from git_system_follower.typings.package import PackageLocalData


__all__ = ['print_params', 'print_list', 'banner', 'print_dependency_tree_one_level']


WIDTH = 100


def print_params(
        params: Mapping, title='', width: int = WIDTH,
        hidden_params: Iterable = (), *,
        output_func: Callable = print
) -> None:
    """ Print parameters as "window" from value as dict

    :param params: value as dict: key/value
    :param title: window's title
    :param width: parameters window width
    :param hidden_params: list of <params> keys to hide
    :param output_func: output function
    """
    width = width - 2  # subtract the sides
    max_length_param = len(max(params.keys(), key=lambda x: len(x)))

    content = f'\n{_get_header(title, width=width)}\n'
    for key, value in params.items():
        param = key.ljust(max_length_param)
        if key in hidden_params:
            value = '*****' if value else ''
        content += f'  {param} = {value}\n'
    content += f"╰{'═' * width}╯"
    output_func(content)


def _get_header(title: str, *, width: int) -> str:
    title = f' {title} ' if title != '' else ''
    header = title.center(width, '═')
    return f'╭{header}╮'


def print_list(
        elements: Sequence, title: str, width: int = WIDTH, *,
        key: Callable, output_func: Callable = print
) -> None:
    """ Print the list by filtering the information from the list using the function

    :param elements: value as dict: key/value
    :param title: list's title
    :param width: text content width
    :param key: function for filtering the information from the list
    :param output_func: output function
    (e.g. key=lambda package: f"{package['name']}@{package['version']}")
    """
    content = f'{title} ({len(elements)})'
    if len(elements) != 0:
        elements_content = '  '.join([key(elem) for elem in elements])
        content += f':\n{textwrap.fill(elements_content, width)}'
    output_func(content)


def banner(version: str, *, output_func: Callable = print):
    # This logo is colored #F4511E
    content = f"""
    \033[38;2;244;81;30m.-,\033[0m
 \033[38;2;244;81;30m.^.: :.^.\033[0m   ┏┓╻┳ ┏┓╻╻┏┓┳┏┓┏┳┓ ┏┓┏┓╻ ╻ ┏┓┏ ┓┏┓┳┓
\033[38;2;244;81;30m,-' .-. '-,\033[0m  ┃┓┃┃ ┗┓┗┃┗┓┃┣ ┃┃┃ ┣ ┃┃┃ ┃ ┃┃┃┃┃┣ ┣┛
\033[38;2;244;81;30m'-. '-' .-'\033[0m  ┗┛╹╹ ┗┛┗┛┗┛╹┗┛╹ ╹ ╹ ┗┛┗┛┗┛┗┛┗┻┛┗┛┛┗
 \033[38;2;244;81;30m'.`; ;`.'\033[0m   git-system-follower v{version}
    \033[38;2;244;81;30m`-`\033[0m"""
    output_func(content)


def print_dependency_tree_one_level(
        packages: Iterable[PackageLocalData], title='', *,
        key: Callable, output_func: Callable = print
) -> None:
    """ Print dependency tree

    :param packages: packages which need to print
    :param title: title of tree
    :param key: function for filtering the information from the list
    :param output_func: output function
    """
    content = f'{title}:\n'
    for i, package in enumerate(packages, 1):
        content += f'{i}. {key(package)}\n'
        prefix = ' ' * len(str(i)) + '  '  # spaces before connector to level the tree
        for j, dependency in enumerate(package['dependencies']):
            connector = '└── ' if j == len(package['dependencies']) - 1 else '├── '
            content += f'{prefix}{connector}{dependency}\n'

    if content[-1] == '\n':
        content = content[:-1]
    output_func(content)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api to work with CI/CD variables """
from typing import TypedDict

from gitlab.v4.objects import Project
from gitlab.base import RESTObject
from gitlab.exceptions import GitlabCreateError

from git_system_follower.logger import logger


__all__ = ['CICDVariable', 'get_cicd_variables', 'create_variable', 'delete_variable']


class CICDVariable(TypedDict):
    """ Gitlab CI/CD variables

    Dictionary items:
        - name : str | variable name
        - value : str | variable value
        - env : str | Gitlab environment in which to create variable
        - masked : bool | whether to mask variables in log messages (in Gitlab, in git-system-follower log messages)
    """
    name: str
    value: str
    env: str
    masked: bool


def get_cicd_variables(project: Project) -> dict[str, CICDVariable]:
    """ Get CI/CD variables from remote repository

    :param project: GitLab project

    :return: CI/CD variables dict: key - variable name, value - variable information
    """
    gitlab_variables = project.variables.list()
    variables = {}
    for variable in gitlab_variables:
        variables[variable.key] = CICDVariable(
            name=variable.key, value=variable.value, env=variable.environment_scope, masked=variable.masked
        )
    return variables


def create_variable(
        project: Project, variable: CICDVariable, *,
        is_force: bool
) -> RESTObject | None:
    """ Creating CI/CD variable

    :param project: Gitlab project
    :param variable: CI/CD variable to be created
    :param is_force: forced creation
    :return: Gitlab REST API response
    """
    logger.info('\t-> Creating CI/CD variables')
    is_masked = variable['masked']
    masked_value = '*****' if is_masked else variable['value']

    remote_variables: list[RESTObject] = project.variables.list()
    for remote_var in remote_variables:
        remote_masked_value = '*****' if remote_var.masked else remote_var.value
        if variable['name'] != remote_var.key:
            continue

        if variable['value'] == remote_var.value:
            logger.info(f"\t\tVariable with this name ({variable['name']}) and "
                        f"value ({masked_value}) already exist. Skip creation")
            return remote_var
        if is_force:
            remote_var.value = variable['value']
            remote_var.save()
            logger.warning(f"\t\tVariable with this name ({variable['name']}) already exist with different value "
                           f"(old value: {remote_masked_value}, new value: {masked_value}). "
                           f"Force flag enabled. Overwrote CI/CD variable")
            return remote_var
        logger.warning(f"\t\tVariable with this name ({variable['name']}) already exist with different value "
                       f"(old value: {remote_masked_value}, new value: {masked_value}). "
                       f"Force flag disabled. Skip creation/update")
        return
    try:
        variable = project.variables.create({
            'key': variable['name'], 'value': variable['value'], 'environment_scope': variable['env'],
            'masked': variable['masked']
        })
    except GitlabCreateError:
        msg = (
            f"Failed to create {variable['name']} CI/CD variable with {masked_value} value. "
            f"Please make sure you follow Gitlab's rules for CI/CD variables "
            f"https://docs.gitlab.com/ee/ci/variables/#add-a-cicd-variable-to-a-project."
        )
        if is_masked:
            msg += (" You are using variable masking. Please make sure you follow Gitlab's rules of variable masking "
                    "https://docs.gitlab.com/ee/ci/variables/#mask-a-cicd-variable")
        logger.critical(msg)
        raise
    logger.info(f"\t\tVariable with name {variable.key} with value {masked_value} has been created "
                f"in {variable.environment_scope} environment")
    logger.debug(f'\t\tResponse:\n{variable.pformat()}')
    return variable


def delete_variable(
        project: Project, variable: CICDVariable, *,
        is_force: bool
) -> None:
    """ Deleting CI/CD variable

    :param project: Gitlab project
    :param variable: CI/CD variable to be deleted
    :param is_force: forced deletion
    """
    logger.info('\t-> Deleting CI/CD variables')
    masked_value = '*****' if variable['masked'] else variable['value']

    remote_variables: list[RESTObject] = project.variables.list()
    for remote_var in remote_variables:
        remote_masked_value = '*****' if remote_var.masked else remote_var.value
        if variable['name'] != remote_var.key:
            continue

        if variable['value'] == remote_var.value:
            project.variables.delete(variable['name'])
            logger.info(f"\t\tVariable with this name ({variable['name']}) and "
                        f"value ({masked_value}) exist. Deleted it")
            return
        if is_force:
            project.variables.delete(variable['name'])
            logger.warning(f"\t\tVariable with this name ({variable['name']}) exist with different value "
                           f"(old value: {remote_masked_value}, new value: {masked_value}). "
                           f"Force flag enabled. Deleted CI/CD variable")
            return
        logger.warning(f"\t\tVariable with this name ({variable['name']}) exist with different value "
                       f"(old value: {remote_masked_value}, new value: {masked_value}). "
                       f"Force flag disabled. Skip deletion")
        return
    logger.info(f"\t\tNot found CI/CD variable with {variable['name']} name for deletion")



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Package description file (package.yaml) processing module """
from pathlib import Path
from typing import TypedDict, Any

import yaml

from git_system_follower.logger import logger
from git_system_follower.errors import (
    PackageNotFoundError, PackageDescriptionFileError, DescriptionSectionError, MaxDependencyDepthError
)
from git_system_follower.variables import PACKAGE_DIRNAME, PACKAGE_DESCRIPTION_FILE_API
from git_system_follower.plugins.cli.packages.default import ImagePlugin
from git_system_follower.typings.package import PackageData, PackageLocalData


__all__ = [
    'DESCRIPTION_FILENAME',
    'get_package_info', 'add_dependencies', 'check_dependency_depth'
]


DESCRIPTION_FILENAME = 'package.yaml'
MAX_DEPENDENCY_LEVEL = 1
PACKAGE_DESCRIPTION_FILE_API: dict[str, 'ApiVersionInfo']


class ApiVersionInfo(TypedDict):
    mandatory_sections: tuple[str, ...]
    optional_sections: tuple[str, ...]
    section_types: tuple[object, ...]
    package_types: tuple[str, ...]


def _has_downloaded(path: Path) -> bool:
    """ Has package been downloaded

    :param path: path to directory with local package
    :return: True/False
    """
    return path.exists()


def add_dependencies(
        packages: list[PackageLocalData], dependencies: list[PackageLocalData],
        is_deps_first: bool
) -> list[PackageLocalData]:
    for dependency in dependencies:
        if dependency in packages:
            continue

        if is_deps_first:
            packages.insert(-1, dependency)
        else:
            packages.append(dependency)
    return packages


def get_package_info(directory: Path, name: str) -> PackageLocalData:
    """ Get package info from package description file (package.yaml)

    :param directory: path to directory with local package
    :param name: package name

    :return: local package info
    """
    path = directory / PACKAGE_DIRNAME / DESCRIPTION_FILENAME
    if not directory.exists():
        raise PackageNotFoundError(f'No such directory with {name} package ({directory.absolute()})')
    if not path.parent.exists():
        raise PackageNotFoundError(f'No such directory inside the project with package files {path.parent.absolute()}')
    if not path.exists():
        raise PackageNotFoundError(f'No such package.yaml file for {name} package ({path.absolute()})')

    with open(path, 'r', encoding='utf-8') as file:
        data = yaml.safe_load(file)

    try:
        data = _validate_package_info(data)
    except PackageDescriptionFileError:
        logger.critical(f'Error during validation of {DESCRIPTION_FILENAME} file ({path.absolute()})')
        raise

    return PackageLocalData(
        **data,
        path=directory
    )


def _validate_package_info(data: dict[str, Any]) -> PackageData:
    """ Validate package information

    :param data: information from package description file (package.yaml)
    :return: ready-made package information
    """
    # TODO: implement package.yaml validate
    api_version = _validate_api_version_section(data)
    api_info = PACKAGE_DESCRIPTION_FILE_API[api_version]
    _validate_section_names(data, api_info['mandatory_sections'], api_info['optional_sections'])
    # _validate_section_types(data, api_info['sections'], api_info['section_types'])
    _validate_type_section(data['type'], api_info['package_types'])
    updated_dependencies = []
    dependencies = data.get('dependencies')
    if dependencies is not None:
        image_plugin = ImagePlugin()
        for dependency in data['dependencies']:
            dependency = image_plugin.parse_image(dependency)
            updated_dependencies.append(dependency)
    data['version'] = str(data['version'])
    data['dependencies'] = tuple(updated_dependencies)
    return data


def _validate_api_version_section(data: dict[str, Any]) -> str:
    if (api_version := data.get('apiVersion')) is None:
        raise DescriptionSectionError("Section 'apiVersion' missing")

    if api_version not in PACKAGE_DESCRIPTION_FILE_API.keys():
        raise DescriptionSectionError(f"Unsupported package description file api version: '{api_version}'. "
                                      f"Available versions: {', '.join(PACKAGE_DESCRIPTION_FILE_API.keys())}")
    return api_version


def _validate_section_names(
        data: dict[str, Any], mandatory_keys: tuple[str, ...], optional_keys: tuple[str, ...]
) -> None:
    """ Mandatory keys validation

    :param data: information from package description file (package.yaml)
    """
    for mandatory_key in mandatory_keys:
        if mandatory_key not in data.keys():
            raise DescriptionSectionError(f'No mandatory section in {DESCRIPTION_FILENAME} file ({mandatory_key})')

    for key in data.keys():
        if key not in mandatory_keys and key not in optional_keys:
            raise DescriptionSectionError(f'An extra section is specified in {DESCRIPTION_FILENAME} file ({key})')


def _validate_section_types(data: dict[str, Any], sections: tuple[str, ...], types: tuple[object, ...]) -> None:
    # TODO: implement package.yaml validate: section types
    pass


def _validate_type_section(package_type: str, available_types: tuple[str, ...]) -> None:
    if package_type not in available_types:
        raise DescriptionSectionError(f"Unsupported package type: '{package_type}'. "
                                      f"Available types: {', '.join(available_types)}")


def check_dependency_depth(level: int, tree: str) -> None:
    """ Checking the dependency tree for maximum depth

    :param level: current dependency depth level
    :param tree: current dependency tree/string to printing the problematic dependency chain (to raise error),
                 e.g. `parent-package -> parent-dependency -> problem-dependency`
    """
    if level > MAX_DEPENDENCY_LEVEL:
        msg = f'The maximum dependency level has been reached ({MAX_DEPENDENCY_LEVEL}). Error for {tree}'
        raise MaxDependencyDepthError(msg)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from sys import stdin
import base64
import binascii
import os

import click

from git_system_follower.logger import logger
from git_system_follower.typings.cli import PackageCLISource, PackageCLITarGz, PackageCLIImage, ExtraParam, Credentials
from git_system_follower.plugins.managers import cli_packages_pm as plugin_manager
from git_system_follower.plugins.cli.packages.specs import HookSpec


__all__ = [
    'Package', 'PackageType', 'ExtraParamTuple',
    'resolve_credentials',
    'add_options', 'get_gears'
]


class PackageType(click.ParamType):
    """ Class for checking parameters from click cli """
    name = 'package'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value: str, param, ctx) -> PackageCLIImage | PackageCLITarGz | PackageCLISource:
        return plugin_manager.process(value, **ctx.params)


Package = PackageType()


class ExtraParamTuple(click.Tuple):
    name = 'extra_param'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def convert(self, value, param, ctx):
        values = super().convert(value, param, ctx)
        return ExtraParam(name=values[0], value=values[1], masked=True if values[2] == 'masked' else False)


def resolve_credentials(cli_username: str | None, cli_password: str | None) -> Credentials | None:
    """ Resolve credentials obtained in different ways

    Priorities:
        1. cli parameters
        2. stdin/pipe parameters (parse to first ":")
        3. env variables
        4. prompts if only username or only password is specified
    """
    stdin_username, stdin_password = read_stdin_credentials()

    username = cli_username or stdin_username or os.getenv("GSF_REGISTRY_USERNAME")
    password = cli_password or stdin_password or os.getenv("GSF_REGISTRY_PASSWORD")

    if username is None and password is None:
        logger.debug('Username and/or password have not been passed by either method')
        return None

    if username is None:
        username = click.prompt('Registry username', type=str)
    if password is None:
        password = click.prompt('Registry password', type=str, hide_input=True)
    return Credentials(username=username, password=password)


def read_stdin_credentials() -> tuple[Optional[str], Optional[str]]:
    """ Reads username and password from standard input.

    :returns: Credentials or None if stdin is empty
    """
    if stdin.isatty():
        return None, None

    creds = stdin.read().strip()
    logger.debug('Encountered a string in stdin. Trying to parse it')
    if ':' not in creds:
        logger.debug('No ":" separator found, attempt to unmask string with base64')
        try:
            creds = base64.b64decode(creds).decode().strip()
        except (binascii.Error, UnicodeDecodeError) as error:
            logger.warning(
                f'The separator ":" in the stdin string was not found. '
                f'When trying to unmask this string using base64, we got an error: {error}. '
                f'Skip parse stdin string'
            )
            return None, None
    if ':' not in creds:
        return None, creds
    username, password = creds.split(':', 1)
    return username, password


""" --------------- For plugins --------------- """


def get_gears(hooks: tuple[HookSpec, ...]) -> tuple[PackageCLISource | PackageCLITarGz | PackageCLIImage, ...]:
    """ Add gears from plugin """
    gears = []
    for hook in hooks:
        gears.extend(hook.gears)
    return tuple(gears)


def add_options(command: object, managers: list):
    """ Dynamically add options to click command """
    for manager in managers:
        options = manager.get_plugin_options()
        for plugin_name, opts in options.items():
            for opt in opts:
                params = extract_click_option_params(opt)

                help_msg = params.get('help', '')
                params['help'] = f'{manager.group}:{plugin_name}:{help_msg}'
                command = opt(command)

    return command


def extract_click_option_params(option: click.option) -> dict:
    """ Extract click.option params """
    closure = option.__closure__
    if not closure:
        return {}

    for cell in closure:
        if isinstance(cell.cell_contents, dict):
            return cell.cell_contents

    return {}



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable
import functools


class NeedRetry(Exception):
    def __init__(self, msg='', *args):
        super().__init__(msg, *args)


class MaxRetries(Exception):
    def __init__(self, msg, *args):
        super().__init__(msg, *args)


def retry(max_retries: int = 5, *, output_func: Callable = print, error_output_func: Callable = print):
    """ Decorator to re-execute the function if an exception occurs

    :param max_retries: Maximum number of attempts to execute the function
    :param output_func: normal message output function
    :param error_output_func: error output function
    """
    def decorator_retry(func):
        @functools.wraps(func)
        def wrapper_retry(*args, **kwargs):
            count = 0
            while count < max_retries:
                try:
                    result = func(*args, **kwargs)
                    return result
                except NeedRetry as error:
                    count += 1
                    error_output_func(f'{error.__class__.__name__}: {error}')

                if count >= max_retries:
                    raise MaxRetries('Max retries reached. Operation failed')
                output_func(f'There is been an error. Retry №{count}')
        return wrapper_retry
    return decorator_retry



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import Path

import click
from colorama import Fore, Style


# for stdout
_SHORT_FORMAT = '[%(asctime)s.%(msecs)03d] %(levelname)-18s | %(message)s'
_SHORT_DATE_FORMAT = '%H:%M:%S'

# for log file
_FORMAT = '[%(asctime)s.%(msecs)03d] %(levelname)-8s | %(filename)s:%(funcName)s:%(lineno)d - %(message)s'
_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

# New logging level
SUCCESS_LEVEL_NUM = 25
SUCCESS_LEVEL_NAME = 'SUCCESS'


class ColoredFormatter(logging.Formatter):

    COLORS = {
        'DEBUG': Fore.BLUE,
        'INFO': Fore.WHITE,
        'SUCCESS': Fore.GREEN,
        'WARNING': Fore.YELLOW,
        'ERROR': Fore.RED,
        'CRITICAL': f'{Fore.RED}{Style.BRIGHT}',
    }

    def format(self, record):
        color = self.COLORS.get(record.levelname, '')
        if color:
            record.name = f'{color}{record.name}{Fore.WHITE}'
            record.levelname = f'{color}{record.levelname}{Fore.WHITE}'
            record.msg = f'{color}{record.msg}{Fore.WHITE}'
        return logging.Formatter.format(self, record)


class RemoveColorFilter(logging.Filter):
    # for file handlers
    def filter(self, record):
        if record and record.msg and isinstance(record.msg, str):
            record.levelname = click.unstyle(record.levelname)
            record.msg = click.unstyle(record.msg)
        return True


file_formatter = logging.Formatter(fmt=_FORMAT, datefmt=_DATE_FORMAT)


def disable_info_for_other_loggers(names: list[str]) -> None:
    for name in names:
        logger = logging.getLogger(name)
        logger.setLevel(logging.WARNING)


def get_stream_handler() -> logging.StreamHandler:
    handler = logging.StreamHandler()
    handler.setLevel(level=logging.INFO)
    handler.setFormatter(ColoredFormatter(fmt=_SHORT_FORMAT, datefmt=_SHORT_DATE_FORMAT))
    return handler


def get_file_handler(path: Path) -> logging.FileHandler:
    path.parent.mkdir(parents=True, exist_ok=True)
    handler = logging.FileHandler(path, mode='w', encoding='utf-8')
    handler.setLevel(level=logging.DEBUG)
    handler.addFilter(RemoveColorFilter())
    handler.setFormatter(file_formatter)
    return handler


def success(self, message, *args, **kwargs):
    if self.isEnabledFor(SUCCESS_LEVEL_NUM):
        self._log(SUCCESS_LEVEL_NUM, message, args, **kwargs)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for working with GitLab REST API """
from urllib.parse import urlparse
from time import sleep
from pprint import pformat

from gitlab import Gitlab
from gitlab.v4.objects import Project, ProjectMergeRequest
import gitlab.exceptions

from git_system_follower.logger import logger
from git_system_follower.errors import RemoteRepositoryError, HashesMismatch
from git_system_follower.states import StateFile
from git_system_follower.package.cicd_variables import get_cicd_variables
from git_system_follower.utils.retry import NeedRetry


__all__ = ['get_gitlab', 'get_project', 'get_states', 'create_mr', 'merge_mr']


# variables for waiting for an update in a remote repository
WAIT = 4
MAX_WAIT = 60


def get_gitlab(url: str, token: str) -> Gitlab:
    """ Get gitlab instance for working with GitLab REST API
    :param url: any gitlab url
    :param token: gitlab access token
    """
    url = _shorten_url(url)
    instance = Gitlab(url, private_token=token)
    instance.auth()
    return instance


def _shorten_url(url: str) -> str:
    parsed = urlparse(url)
    return f'{parsed.scheme}://{parsed.netloc}'


def get_project(instance: Gitlab, url: str) -> Project:
    project_with_namespace = urlparse(url).path[1:].replace('.git', '')
    try:
        project = instance.projects.get(project_with_namespace)
    except gitlab.exceptions.GitlabGetError:
        raise RemoteRepositoryError(f'Project/repository {url} not found')
    except gitlab.exceptions.GitlabAuthenticationError:
        raise RemoteRepositoryError(f'Failed to auth in {url} repository')
    return project


def get_states(project: Project, branches: tuple[str, ...]) -> dict[str, StateFile]:
    """ Get states files using GitLab REST API

    :param project: GitLab project
    :param branches: branch names listing

    :return: return dictionary with key - branch name, value - state file for this branch
    """
    states = {}
    remote_branches = [branch.name for branch in project.branches.list(get_all=True)]
    cicd_variables = get_cicd_variables(project)
    for branch in branches:
        if branch not in remote_branches:
            raise RemoteRepositoryError(f'Branch {branch} not found')

        try:
            raw = project.files.raw(file_path='.state.yaml', ref=branch)
            states[branch] = StateFile(raw=raw, current_cicd_variables=cicd_variables)
        except gitlab.exceptions.GitlabGetError:
            states[branch] = StateFile()
        except HashesMismatch as error:
            logger.critical(f'Hashes do not match for {branch} branch. Most likely, someone changed the state file '
                            f'manually, this is forbidden by package manager policy. Please reset everything back to '
                            f'its original state and start again. '
                            f'State file hash: {error.state_file_hash} != {error.generated_hash}: Generated hash')
            raise
    return states


def create_mr(
        project: Project, source: str, target: str, *,
        title: str = 'Install package(s)', description: str = ''
) -> ProjectMergeRequest:
    # Merge Request is auto closed when a branch is deleted
    mr = project.mergerequests.create({
        'source_branch': source,
        'target_branch': target,
        'title': title,
        'description': description,
        'squash': True,
        'remove_source_branch': True
    })
    logger.success(f'Created merge requests {source} -> {target} (url: {mr.web_url})')
    logger.debug(f'Response:\n{mr.pformat()}')
    mr = project.mergerequests.get(mr.iid)
    return mr


def merge_mr(project: Project, mr: ProjectMergeRequest) -> dict:
    total = 0
    while mr.merge_status == 'checking':
        logger.debug(f'Waiting to be able to merge ({WAIT} sec)')
        sleep(WAIT)
        mr = project.mergerequests.get(mr.iid)
        total += WAIT
        if total > MAX_WAIT:
            raise RemoteRepositoryError(f'Waiting too long for a merger opportunity ({MAX_WAIT} sec)')

    if mr.has_conflicts:
        raise NeedRetry(f'Cannot merge {mr.source_branch} -> {mr.target_branch} because there are conflicts')
    response = mr.merge()
    logger.success(f'Merged {mr.source_branch} -> {mr.target_branch} (url: {mr.web_url})')
    logger.debug(f'Response:\n{pformat(response)}')
    return response



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import re

from git_system_follower.errors import ParsePackageNameError
from git_system_follower.typings.cli import PackageCLISource, PackageCLITarGz, PackageCLIImage
from git_system_follower.plugins.cli.packages import hookimpl
from git_system_follower.plugins.cli.packages.specs import HookSpec


__all__ = ['SourcePlugin', 'TarGzPlugin', 'ImagePlugin']


class SourcePlugin(HookSpec):
    @hookimpl
    def match(self, value: str) -> bool:
        path = Path(value)
        return path.is_dir()

    @hookimpl
    def get_gears(self, value: str, **kwargs) -> list[PackageCLISource]:
        return [PackageCLISource(path=Path(value))]

    def __str__(self) -> str:
        return self.value


class TarGzPlugin(HookSpec):
    suffix = '.tar.gz'

    @hookimpl
    def match(self, value: str) -> bool:
        path = Path(value)
        return path.name.endswith(self.suffix)

    @hookimpl
    def get_gears(self, value: str, **kwargs) -> list[PackageCLITarGz]:
        return [PackageCLITarGz(path=Path(value))]

    def __str__(self) -> str:
        return self.value


class ImagePlugin(HookSpec):
    pattern = (
        r'^(?P<registry>[^:/]+(?::\d+)?)\/'
        r'(?:(?P<path>[^/]+(?:\/[^/]+)*)\/)?'
        r'(?P<image_name>[^:\/]+)'
        r'(?::(?P<image_version>.+))?$'
    )

    @hookimpl
    def match(self, value: str) -> bool:
        if re.match(self.pattern, value):
            return True
        return False

    @hookimpl
    def get_gears(self, value: str, **kwargs) -> list[PackageCLIImage]:
        return [self.parse_image(value)]

    def parse_image(self, package: str) -> PackageCLIImage:
        match = re.match(self.pattern, package)
        if not match:
            raise ParsePackageNameError(f'Failed to parse {package} package name with regular expression')

        registry, repository = match.group('registry'), match.group('path')
        image, tag = match.group('image_name'), match.group('image_version')
        if tag is None:
            return PackageCLIImage(registry=registry, repository=repository, image=image)
        return PackageCLIImage(registry=registry, repository=repository, image=image, tag=tag)

    def __str__(self) -> str:
        return self.value



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.metadata

import pluggy
import click

from git_system_follower.logger import logger
from git_system_follower.errors import ParsePackageNameError, InvalidPlugin
from git_system_follower.plugins.cli.packages import NAME
from git_system_follower.plugins.cli.packages.specs import HookSpec
from git_system_follower.plugins.cli.packages.default import SourcePlugin, TarGzPlugin, ImagePlugin
from git_system_follower.utils.output import print_list


def validate_hooks(method):
    def wrapper(self, plugin):
        requires = ['match', 'process', 'get_gears']
        not_exist = []
        for require in requires:
            if require not in dir(plugin):
                not_exist.append(require)

        if not_exist:
            raise InvalidPlugin(f'Incorrect plugin structure. Missing hooks: {", ".join(not_exist)}')
        return method(self, plugin=plugin)
    return wrapper


class PluginManager:
    group = 'gsf.plugins.cli.packages'

    def __init__(self):
        self.pm = pluggy.PluginManager(NAME)
        self.pm.add_hookspecs(HookSpec)
        self.load_plugins()

    def load_plugins(self) -> list[object]:
        """ Load user's plugins from entry points

        User plugins are loaded first, and only then system plugins. To be able to override the default behavior
        """
        plugins = self._load_entry_points_plugins()

        system_plugins = [ImagePlugin, TarGzPlugin, SourcePlugin]
        for plugin in system_plugins:
            self.register(plugin())
            plugins.append(plugin)

        print_list(
            plugins,
            f'Loaded plugins for input package processing ({self.group})',
            key=lambda p: p.__name__, output_func=logger.debug
        )
        return plugins

    def _load_entry_points_plugins(self) -> list[object]:
        plugins = []

        entry_points = importlib.metadata.entry_points()
        # fix for python < 3.11
        if isinstance(entry_points, dict):
            entry_points = entry_points.get(self.group, [])

        for entry_point in entry_points:
            if entry_point.group == self.group:
                plugin = entry_point.load()
                self.register(plugin())
                plugins.append(plugin)
        return plugins

    @validate_hooks
    def register(self, plugin) -> None:
        self.pm.register(plugin)

    def get_plugin_options(self) -> dict[str, list[click.Option]]:
        options = {}
        for hook in self.pm.hook.plugin_options.get_hookimpls():
            opts = hook.plugin.plugin_options()
            options[hook.plugin.__class__.__name__] = opts
        return options

    def process(self, value: str, **kwargs) -> HookSpec:
        """ Processing input package value from CLI
        if no hook implementation has processed package then raise error

        :param value: package string for processing
        :param kwargs: plugin's parameters
        :return: processed package
        """
        for hook in self.pm.hook.match.get_hookimpls():
            process = hook.plugin.process
            is_processed = process(value=value, **kwargs)
            if is_processed:
                return hook.plugin

        # fix to get plugins in the correct order, not self.pm.get_plugins()
        plugins = [plugin[1].__class__.__name__ for plugin in self.pm.list_name_plugin()]
        raise ParsePackageNameError(
            f'Failed to determine package type of "{value}". Available system types: docker image, '
            f'local .tar.gz archive, local source directory. All plugins: {", ".join(plugins)}. '
            f'If you specified an .tar.gz archive or directory, please make sure it exist'
        )



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import click

from git_system_follower.logger import logger, set_level
from git_system_follower.errors import CLIParamsError
from git_system_follower.plugins.cli.packages.specs import HookSpec
from git_system_follower.typings.cli import ExtraParam
from git_system_follower.typings.registry import RegistryTypes, RegistryInfo
from git_system_follower.utils.cli import Package, ExtraParamTuple, resolve_credentials, add_options, get_gears
from git_system_follower.utils.output import banner, print_params
from git_system_follower.git_api.utils import get_config
from git_system_follower.download import download
from git_system_follower.install import install
from git_system_follower.uninstall import uninstall
from git_system_follower import __version__

from git_system_follower.plugins.managers import managers


config = get_config('~/.gitconfig')
GIT_USERNAME = config.get_value('user', 'name', default='unknown')
GIT_EMAIL = config.get_value('user', 'email', default='unknown@example.com')


@click.command(name='download')
@click.argument('gears', nargs=-1, type=Package)
@click.option(
    '-d', '--directory', type=click.Path(exists=True, dir_okay=True, file_okay=False, path_type=Path),
    default=Path('.'), help='Directory where gears will be downloaded'
)
@click.option(
    '--registry-type',
    type=click.Choice([registry_type.value for registry_type in RegistryTypes], case_sensitive=False),
    required=False, default='Autodetect',
    help='Specify the registry type or use automatic detection'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-username', type=str, required=False, default=None,
    help='Username for basic authentication in the registry when downloading Gears'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-password', type=str, required=False, default=None,
    help='Password for basic authentication in the registry when downloading Gears'
)
@click.option(
    '--insecure-registry', 'is_insecure', is_flag=True, default=False,
    help='Allow insecure connections to the registry (use HTTP instead of HTTPS)'
)
@click.option('--debug', 'is_debug', is_flag=True, default=False, help='Show debug level messages')
def download_command(
        gears: tuple[HookSpec, ...], directory: Path,
        registry_type: str, registry_username: str | None, registry_password: str | None, is_insecure: bool,
        is_debug: bool,
        *args, **kwargs  # dont delete, these parameters for plugin manager
):
    """ Download gears

    \b
    GEARS                         Download all listed gears as image:
                                  <registry>/<repository>/<name>:<tag>, e.g.
                                  artifactory.company.com/path-to/your-image:1.0.0
    """
    credentials = resolve_credentials(registry_username, registry_password)
    banner(version=__version__, output_func=logger.info)
    print_params({
        'gears': ', '.join([str(gear) for gear in gears]),
        'directory': directory.absolute(),
        'registry-type': registry_type,
        'registry-username': credentials.username if credentials is not None else '',
        'registry-password': credentials.password if credentials is not None else '',
        'insecure-registry': is_insecure,
        'debug': is_debug
    }, 'Start parameters', hidden_params=('token', 'registry-password'), output_func=logger.info)
    if gears == ():
        raise CLIParamsError('Gears for downloading are not specified')
    set_level(is_debug)

    gears = get_gears(gears)
    registry = RegistryInfo(credentials=credentials, type=RegistryTypes(registry_type), is_insecure=is_insecure)
    download(gears, directory, is_deps_first=True, registry=registry)


@click.command(name='install')
@click.argument('gears', nargs=-1, type=Package)
@click.option(
    '-r', '--repo', 'repo', type=str, required=True,
    help='Gitlab repository url', metavar='URL'
)
@click.option(
    '-b', '--branch', 'branches', type=str, required=True, multiple=True,
    help='Branches in which to install the gears', metavar='BRANCH...'
)
@click.option(
    '-t', '--token', type=str, envvar='GSF_GIT_TOKEN', required=True,
    help='Gitlab access token'
)
@click.option(
    '--extra', 'extras', type=ExtraParamTuple([
        click.STRING, click.STRING, click.Choice(['masked', 'no-masked'], case_sensitive=False)
    ]),
    multiple=True, help='Extra parameters to be passed to the package API: variable name, value, masked/no-masked',
    metavar='<NAME VALUE CHOICE>...'
)
@click.option(
    '--message', type=str, default='Installed gear(s)',
    help='Commit message'
)
@click.option(
    '--git-username', 'username', type=str, envvar='GSF_GIT_USERNAME', default=GIT_USERNAME,
    help='Username under which the commit will be made to the repository', metavar='USER'
)
@click.option(
    '--git-email', 'email', type=str, envvar='GSF_GIT_EMAIL', default=GIT_EMAIL,
    help='User email under which the commit will be made to the repository', metavar='EMAIL'
)
@click.option(
    '--registry-type',
    type=click.Choice([registry_type.value for registry_type in RegistryTypes], case_sensitive=False),
    required=False, default='Autodetect',
    help='Specify the registry type or use automatic detection'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-username', type=str, required=False, default=None,
    help='Username for basic authentication in the registry when downloading Gears'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-password', type=str, required=False, default=None,
    help='Password for basic authentication in the registry when downloading Gears'
)
@click.option(
    '--insecure-registry', 'is_insecure', is_flag=True, default=False,
    help='Allow insecure connections to the registry (use HTTP instead of HTTPS)'
)
@click.option(
    '-f', '--force', 'is_force', is_flag=True, default=False,
    help='Forced installation: change of files, CI/CD variables as specified in gear'
)
@click.option('--debug', 'is_debug', is_flag=True, default=False, help='Show debug level messages')
def install_command(
        gears: tuple[HookSpec, ...], repo: str,
        branches: tuple[str, ...], token: str, extras: tuple[ExtraParam],
        message: str, username: str, email: str,
        registry_type: str, registry_username: str | None, registry_password: str | None, is_insecure: bool,
        is_force: bool, is_debug: bool,
        *args, **kwargs  # dont delete, these parameters for plugin manager
):
    """ Install gears to branches in repository

    \b
    GEARS                         Install all listed gears as
                                  1. image: <registry>/<repository>/<name>:<tag>, e.g.
                                  artifactory.company.com/path-to/your-image:1.0.0
                                  2. .tar.gz archive: /path/to/archive.tar.gz, e.g.
                                  your-archive@1.0.0.tar.gz
                                  3. source code files: /path/to/gear directory, e.g.
                                  your-gear@1.0.0
    """
    credentials = resolve_credentials(registry_username, registry_password)
    banner(version=__version__, output_func=logger.info)
    print_params({
        'gears': ', '.join([str(gear) for gear in gears]),
        'repo': repo,
        'branches': ', '.join(branches),
        'token': token,
        'extras': ', '.join([f"{var.name}={'*****' if var.masked else var.value}" for var in extras]),
        'message': message,
        'git-username': username,
        'git-email': email,
        'registry-type': registry_type,
        'registry-username': credentials.username if credentials is not None else '',
        'registry-password': credentials.password if credentials is not None else '',
        'insecure-registry': is_insecure,
        'force': is_force,
        'debug': is_debug
    }, 'Start parameters', hidden_params=('token', 'registry-password'), output_func=logger.info)
    if gears == ():
        raise CLIParamsError('Gears for installation are not specified')
    set_level(is_debug)

    gears = get_gears(gears)
    registry = RegistryInfo(credentials=credentials, type=RegistryTypes(registry_type), is_insecure=is_insecure)
    install(
        gears, repo, branches, token, extras=extras,
        commit_message=message, username=username, user_email=email,
        registry=registry, is_force=is_force
    )


@click.command(name='uninstall')
@click.argument('gears', nargs=-1, type=Package)
@click.option(
    '-r', '--repo', 'repo', type=str, required=True,
    help='Gitlab repository url', metavar='URL'
)
@click.option(
    '-b', '--branch', 'branches', type=str, required=True, multiple=True,
    help='Branches in which to uninstall the gears', metavar='BRANCH...'
)
@click.option(
    '-t', '--token', type=str, envvar='GSF_GIT_TOKEN', required=True,
    help='Gitlab access token'
)
@click.option(
    '--extra', 'extras', type=ExtraParamTuple([
        click.STRING, click.STRING, click.Choice(['masked', 'no-masked'], case_sensitive=False)
    ]),
    multiple=True, help='Extra parameters to be passed to the package API: variable name, value, masked/no-masked',
    metavar='<NAME VALUE CHOICE>...'
)
@click.option(
    '--message', type=str, default='Uninstalled gear(s)',
    help='Commit message'
)
@click.option(
    '--git-username', 'username', type=str, envvar='GSF_GIT_USERNAME', default=GIT_USERNAME,
    help='Username under which the commit will be made to the repository', metavar='USER'
)
@click.option(
    '--git-email', 'email', type=str, envvar='GSF_GIT_EMAIL', default=GIT_EMAIL,
    help='User email under which the commit will be made to the repository', metavar='EMAIL'
)
@click.option(
    '--registry-type',
    type=click.Choice([registry_type.value for registry_type in RegistryTypes], case_sensitive=False),
    required=False, default='Autodetect',
    help='Specify the registry type or use automatic detection'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-username', type=str, required=False, default=None,
    help='Username for basic authentication in the registry when downloading Gears'
)
@click.option(
    # env variable is specified in resolve_credentials because of priority
    '--registry-password', type=str, required=False, default=None,
    help='Password for basic authentication in the registry when downloading Gears'
)
@click.option(
    '--insecure-registry', 'is_insecure', is_flag=True, default=False,
    help='Allow insecure connections to the registry (use HTTP instead of HTTPS)'
)
@click.option(
    '-f', '--force', 'is_force', is_flag=True, default=False,
    help='Forced uninstallation: change of files, CI/CD variables as specified in gear'
)
@click.option('--debug', 'is_debug', is_flag=True, default=False, help='Show debug level messages')
def uninstall_command(
        gears: tuple[HookSpec, ...], repo: str,
        branches: tuple[str, ...], token: str, extras: tuple[ExtraParam, ...],
        message: str, username: str, email: str,
        registry_type: str, registry_username: str | None, registry_password: str | None, is_insecure: bool,
        is_force: bool, is_debug: bool,
        *args, **kwargs  # dont delete, these parameters for plugin manager
):
    """ Uninstall gears from branches in repository

    It is necessary to have gears, since the manager interacts with the delete package api

    \b
    GEARS                         Uninstall all listed gears as
                                  1. image: <registry>/<repository>/<name>:<tag>, e.g.
                                  artifactory.company.com/path-to/your-image:1.0.0
                                  2. .tar.gz archive: /path/to/archive.tar.gz, e.g.
                                  your-archive@1.0.0.tar.gz
                                  3. source code files: /path/to/gear directory, e.g.
                                  your-gear@1.0.0
    """
    credentials = resolve_credentials(registry_username, registry_password)
    banner(version=__version__, output_func=logger.info)
    print_params({
        'gears': ', '.join([str(gear) for gear in gears]),
        'repo': repo,
        'branches': ', '.join(branches),
        'token': token,
        'extras': ', '.join([f"{var.name}={'*****' if var.masked else var.value}" for var in extras]),
        'message': message,
        'git-username': username,
        'git-email': email,
        'registry-type': registry_type,
        'registry-username': credentials.username if credentials is not None else '',
        'registry-password': credentials.password if credentials is not None else '',
        'insecure-registry': is_insecure,
        'force': is_force,
        'debug': is_debug
    }, 'Start parameters', hidden_params=('token', 'registry-password'), output_func=logger.info)
    if gears == ():
        raise CLIParamsError('Gears for uninstallation are not specified')
    set_level(is_debug)

    gears = get_gears(gears)
    registry = RegistryInfo(credentials=credentials, type=RegistryTypes(registry_type), is_insecure=is_insecure)
    uninstall(
        gears, repo, branches, token, extras=extras,
        commit_message=message, username=username, user_email=email,
        registry=registry, is_force=is_force
    )


@click.command(name='list')
def list_command():
    """ List installed gears: in develop """
    raise NotImplementedError("In development")


@click.command(name='version')
def version_command():
    """ Show version """
    print(__version__)


@click.group()
@click.version_option(__version__, message='%(version)s')
def cli():
    """ The package manager for Git providers. """


# Dynamic addition so that plugins can add their own parameters
download_command = add_options(download_command, managers)
install_command = add_options(install_command, managers)
uninstall_command = add_options(uninstall_command, managers)

cli.add_command(download_command, name='download')
cli.add_command(install_command, name='install')
cli.add_command(uninstall_command, name='uninstall')
cli.add_command(list_command, name='list')
cli.add_command(version_command, name='version')


if __name__ == '__main__':
    cli(prog_name='gsf', show_default=True)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module for defining system information, such as gitlab domain on which package will be installed """

from urllib.parse import urlparse

from gitlab.v4.objects import Project

from git_system_follower.develop.api.types import System, ExtraParams


__all__ = ['get_system_info']


DOMAIN_VAR_NAME = 'GSF_HOST_DOMAIN'


def get_system_info(project: Project, extras: ExtraParams) -> System:
    """ Get system information

    :param project: gitlab project
    :param extras: extra parameters to be passed to package api
    :return: system information
    """
    if DOMAIN_VAR_NAME in extras.keys():
        domain = extras[DOMAIN_VAR_NAME].value
    else:
        domain = urlparse(project.http_url_to_repo).netloc
    return System(host_domain=domain)



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with package types (package.yaml in package root folder) """
from enum import Enum
from typing import TypedDict
from pathlib import Path
from dataclasses import dataclass

from git_system_follower.typings.cli import (
    PackageCLI, PackageCLIImage, PackageCLITarGz, PackageCLISource
)


__all__ = ['PackageType', 'PackageData', 'PackageLocalData', 'PackagesTo']


class PackageType(Enum):
    gitlab_ci_pipeline = 1


class PackageData(TypedDict):
    """ Class for type hinting for python code: data from package.yaml """
    apiVersion: str
    type: PackageType
    name: str
    version: str

    dependencies: tuple[PackageCLI | PackageCLIImage | PackageCLITarGz | PackageCLISource, ...]


class PackageLocalData(PackageData):
    """ Class for type hinting for python code: data from package.yaml with local data """
    path: Path


@dataclass
class PackagesTo:
    # Packages to be installed. Is set by user using cli
    install: tuple[PackageLocalData, ...]
    # Packages to be deleted. Is set by this tool using state files.
    # It is necessary for case when user want to install a lower package version. New versions know
    # how to install/update/delete old versions, and old versions don't know about new versions
    rollback: tuple[PackageLocalData, ...]



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from git_system_follower.logger import logger
from git_system_follower.typings.repository import RepositoryInfo
from git_system_follower.typings.package import PackageLocalData
from git_system_follower.states import PackageState
from git_system_follower.typings.cli import ExtraParam
from git_system_follower.typings.script import ScriptResponse
from git_system_follower.package.deleter import delete
from git_system_follower.package.initer import init


__all__ = ['rollback']


def rollback(
        package: PackageLocalData, old_package: PackageLocalData, repo: RepositoryInfo, state: PackageState, *,
        created_cicd_variables: tuple[str, ...], extras: tuple[ExtraParam, ...], is_force: bool
) -> ScriptResponse:
    logger.info('==> Package rollback')
    delete(old_package, repo, state, created_cicd_variables=created_cicd_variables, extras=extras, is_force=is_force)
    response = init(package, repo, created_cicd_variables=created_cicd_variables, extras=extras, is_force=is_force)
    return response



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api to work with templates """
from pathlib import Path
import hashlib
import shutil
import os

from cookiecutter.main import cookiecutter

from git_system_follower.logger import logger
from git_system_follower.errors import PackageApiError
from git_system_follower.utils.tmpdir import tempdir


__all__ = ['get_template_names', 'create_template', 'delete_template']


def get_template_names(script_dir: Path) -> tuple[str, ...]:
    """ Get available template names in package api """
    path = script_dir / 'templates'
    if not path.exists():
        raise PackageApiError(f'Template directory is missing. Path: {path}')

    return tuple(template.name for template in path.iterdir() if (path / template).is_dir())


@tempdir
def create_template(
        script_dir: Path, template: str | None, target: Path, *,
        tmpdir: Path, variables: dict[str, str], is_force: bool
) -> None:
    logger.info(f'\t-> Creating project using {template} template')
    if template is None:
        logger.info('\t\tNo template specified. Skip operations')
        return
    path = script_dir / 'templates' / template
    if not path.exists():
        raise PackageApiError(f'Template is missing. Template path: {path}')

    cookiecutter(
        template=str(path), output_dir=tmpdir, no_input=True,
        extra_context=_get_extra_content(target, variables=variables)
    )
    _copy_files(tmpdir / target.name, target, is_force=is_force)
    logger.info(f'\t\tSuccessful use template ({path})')


def _copy_files(source: Path, target: Path, *, is_force: bool) -> None:
    paths = source.glob('**/*')
    for path in paths:
        if path.is_dir():
            continue
        _calculate_hash(path)
        relative_path = path.relative_to(source)
        target_path = target / relative_path
        if not target_path.exists():
            target_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(path, target_path)
            logger.info(f'\t\tFile {relative_path} does not exist. Copied it from template')
            continue

        path_hash, target_path_hash = _calculate_hash(path), _calculate_hash(target_path)
        if path_hash == target_path_hash:
            logger.info(f'\t\tContent of {relative_path} file is same. Skip operations')
            continue

        if is_force:
            shutil.copy(path, target_path)
            logger.warning(f'\t\tFile {relative_path} already exists. Force flag enabled. '
                           f'Overwrote it from template')
            continue
        logger.warning(f'\t\tFile {relative_path} already exists. Cannot copy. Skip operations')


def _calculate_hash(path: Path) -> str:
    """ Calculating file content hash

    :param path: file path
    :return: file content hash
    """
    hash_func = hashlib.sha256()
    with open(path, 'rb') as file:
        hash_func.update(file.read())
    return hash_func.hexdigest()


@tempdir
def delete_template(
        script_dir: Path, template: str | None, target: Path, *,
        variables: dict[str, str], tmpdir: Path, is_force: bool
) -> None:
    logger.info(f'\t-> Deleting project files using {template} template')
    if template is None:
        logger.info('\t\tNo template specified. Skip operations')
        return
    path = script_dir / 'templates' / template
    if not path.exists():
        raise PackageApiError(f'Template is missing. Template path: {path}')

    cookiecutter(
        template=str(path), output_dir=tmpdir, no_input=True,
        extra_context=_get_extra_content(target, variables=variables)
    )
    _delete_files(tmpdir / target.name, target, is_force=is_force)
    logger.info(f'\t\tSuccessful delete files using template ({path})')


def _get_extra_content(target: Path, *, variables: dict[str, str]) -> dict[str, str]:
    content = {'gsf_repository_name': target.name}
    content.update(variables)
    return content


def _delete_files(source: Path, target: Path, *, is_force: bool) -> None:
    paths = source.glob('**/*')
    for path in paths:
        if path.is_dir():
            continue
        relative_path = path.relative_to(source)
        target_path = target / relative_path
        if not target_path.exists():
            logger.info(f'\t\tFile {relative_path} does not exist. Skip operations')
            continue

        path_hash, target_path_hash = _calculate_hash(path), _calculate_hash(target_path)
        if path_hash == target_path_hash:
            os.remove(target_path)
            logger.info(f'\t\tContent of {relative_path} file is same. Deleted this file')
            continue

        if is_force:
            os.remove(target_path)
            logger.warning(f'\t\tFile {relative_path} exists and does not match the template. '
                           f'Overwrote it from template')
            continue
        logger.warning(f'\t\tFile {relative_path} exists and does not match the template. '
                       f'Cannot delete. Skip operations')



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api with types for package developers  """
from dataclasses import dataclass
from pathlib import Path

from gitlab.v4.objects import Project

from git_system_follower.typings.cli import ExtraParam
from git_system_follower.package.cicd_variables import CICDVariable


__all__ = [
    'Parameters', 'SystemParameters', 'System',
    'CICDVariable', 'CICDVariables',
    'ExtraParam', 'ExtraParams'
]


@dataclass(frozen=True)
class SystemParameters:
    project: Project
    created_cicd_vars_names: list[str]
    script_dir: Path
    is_force: bool


@dataclass(frozen=True)
class System:
    host_domain: str


ExtraParamName = str
ExtraParams = dict[ExtraParamName, ExtraParam]

CICDVariableName = str
CICDVariables = dict[CICDVariableName, CICDVariable]


@dataclass(frozen=True)
class Parameters:
    __system_params: SystemParameters
    system: System
    workdir: Path
    extras: ExtraParams
    cicd_variables: CICDVariables
    all_cicd_variables: CICDVariables
    used_template: str
    template_variables: dict[str, str]



# Copyright 2024-2025 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Module with api for `uninstall` command """
from pathlib import Path

from gitlab.v4.objects import Project

from git_system_follower.logger import logger
from git_system_follower.errors import UninstallationError
from git_system_follower.typings.cli import (
    PackageCLI, ExtraParam, PackageCLIImage, PackageCLITarGz, PackageCLISource
)
from git_system_follower.typings.registry import RegistryInfo
from git_system_follower.typings.package import PackageLocalData
from git_system_follower.download import download
from git_system_follower.git_api.gitlab_api import (
    get_gitlab, get_project, get_states, create_mr, merge_mr
)
from git_system_follower.git_api.git_api import checkout_to_new_branch, push_installed_packages
from git_system_follower.git_api.utils import get_packages_str, get_git_repo
from git_system_follower.typings.repository import RepositoryInfo
from git_system_follower.utils.output import print_list
from git_system_follower.utils.retry import retry
from git_system_follower.states import (
    ChangeStatus, PackageState, StateFile,
    get_installed_packages
)
from git_system_follower.package.deleter import delete


__all__ = ['uninstall']


def uninstall(
        packages_cli: tuple[PackageCLIImage | PackageCLITarGz | PackageCLISource, ...],
        repo_url: str, branches: tuple[str, ...], token: str, *,
        extras: tuple[ExtraParam, ...], commit_message: str,
        username: str, user_email: str,
        registry: RegistryInfo, is_force: bool
) -> None:
    gitlab_instance = get_gitlab(repo_url, token)
    project = get_project(gitlab_instance, repo_url)
    states = get_states(project, branches)

    packages = get_packages(packages_cli, states, registry=registry)
    if not packages:
        logger.info('No packages of these versions found in state file')
        return
    print_list(packages, title='Packages', output_func=logger.info,
               key=lambda package: f"{package['name']}@{package['version']}")
    logger.info('Processing branches')
    for i, branch in enumerate(branches, 1):
        logger.info(f'[{i}/{len(branches)}] Processing {branch} branch')
        logger.debug(f'Current state in {branch} branch:\n{states[branch]}')
        validated_packages = validate_packages_dependencies(packages, states[branch])
        print_list(validated_packages, title=f'Uninstallation packages in {branch} branch', output_func=logger.info,
                   key=lambda package: f"{package['name']}@{package['version']}")
        if not validated_packages:
            logger.info(f'There are no packages to delete. Skip deletion for {branch} branch')
            continue
        states[branch] = managing_branch(
            project, branch, token, validated_packages, states[branch], extras=extras,
            commit_message=commit_message, username=username, user_email=user_email, is_force=is_force
        )
    logger.success('Uninstallation complete')


def get_packages(
        packages_cli: tuple[PackageCLIImage | PackageCLITarGz | PackageCLISource, ...],
        states: dict[str, StateFile], *,
        registry: RegistryInfo
) -> tuple[PackageLocalData, ...]:
    installed_packages = get_installed_packages(states)
    downloaded_packages = download(packages_cli, is_deps_first=False, registry=registry)
    packages = []
    for download_package in downloaded_packages:
        for installed_package in installed_packages:
            if _is_necessary_package_to_delete(download_package, installed_package):
                packages.append(download_package)
    return tuple(packages)


def _is_necessary_package_to_delete(download_package: PackageLocalData, installed_package: PackageCLI) -> bool:
    """ Does package need to be deleted: initial check for which packages need to be deleted
    at all (without dependencies)

    :param download_package: downloaded package (package from cli)
    :param installed_package: installed packages from state file
    """
    if download_package['version'] is None:
        return download_package['name'] == installed_package.name
    return (download_package['name'] == installed_package.name and
            download_package['version'] == installed_package.version)


def validate_packages_dependencies(
        packages: tuple[PackageLocalData, ...], state: StateFile
) -> tuple[PackageLocalData, ...]:
    """ Validate packages on which packages should be deleted and which packages should not be deleted.
    To avoid errors when you delete dependencies of one package and the second package uses the same dependency(ies)

    :param packages: packages to be deleted
    :param state: packages specified in branch in state file

    :return: validated packages without dependencies that are used in multiple packages
    """
    result = []
    for i, package in enumerate(packages):
        if _whether_to_delete(package, packages[:i] + packages[i + 1:], state):
            result.append(package)
    return tuple(result)


def _whether_to_delete(
        package: PackageLocalData, packages: tuple[PackageLocalData, ...], state: StateFile
) -> bool:
    """ Does this package need to be removed from the package deletion list

    :param package: package we are looking at
    :param packages: remaining package deletion list (without <package>)
    :param state: current installed package list

    :return: True - if <package> is dependency of another package and this other package should not be deleted;
             False - if <package> is dependency of another package and this other package should be deleted or
                     <package> is not dependency of another package
    """
    is_dependency, for_packages = _is_package_a_dependency(package, state.get_packages())
    if not is_dependency:
        return True
    msg_packages = ', '.join([f"{for_package['name']}@{for_package['version']}" for for_package in for_packages])
    logger.debug(f"{package['name']}@{package['version']} is a dependency for installed packages: {msg_packages}")
    is_delete_main_packages = _whether_to_delete_main_packages(packages, for_packages)

    if is_delete_main_packages:
        logger.debug(f"All packages ({msg_packages}) with {package['name']}@{package['version']} as a dependency "
                     f"will be uninstalled. So uninstall {package['name']}@{package['version']} package")
        return True
    return False


def _is_package_a_dependency(
        package: PackageLocalData, installed_packages: list[PackageState]
) -> tuple[bool, list[PackageState]]:
    """ Whether a package is a dependency of another package

    :param package: package we are looking at
    :param installed_packages: current installed package list

    :return: is dependency of another package; list of another packages which have <package> as dependency
    """
    is_dependency, for_packages = False, []
    for installed_package in installed_packages:
        for dependency in installed_package['dependencies']:
            if dependency == f"{package['name']}@{package['version']}":
                is_dependency = True
                for_packages.append(installed_package)
    return is_dependency, for_packages


def _whether_to_delete_main_packages(
        packages_to_delete: tuple[PackageLocalData, ...], another_packages_with_deps: list[PackageState]
) -> bool:
    """ Whether all packages with dependencies are included in list to be deleted

    :param packages_to_delete: remaining package deletion list (without <package>)
    :param another_packages_with_deps: list of another packages which have <package> as dependency

    :return: whether main packages that contain <package> as a dependency will be deleted
    """
    packages_to_delete_with_deps = [package for package in packages_to_delete if package['dependencies']]

    checked = []
    for another_package_with_deps in another_packages_with_deps:
        for package_to_delete_with_deps in packages_to_delete_with_deps:
            if (another_package_with_deps['name'] == package_to_delete_with_deps['name'] and
                    another_package_with_deps['version'] == package_to_delete_with_deps['version']):
                checked.append(another_package_with_deps)

    for package in another_packages_with_deps:
        if package not in checked:
            logger.debug(f"{package['name']}@{package['version']} with this dependency will not be uninstalled, "
                         f"exclude dependency from uninstallation list")
            break
    return another_packages_with_deps == checked


@retry(output_func=logger.info, error_output_func=logger.error)
def managing_branch(
        project: Project, branch: str, token: str, packages: tuple[PackageLocalData, ...], state: StateFile, *,
        extras: tuple[ExtraParam, ...], commit_message: str, username: str, user_email: str,
        is_force: bool
) -> StateFile:
    repo = RepositoryInfo(gitlab=project, git=get_git_repo(project, token))
    checkout_to_new_branch(repo.git, branch)

    logger.info(':: Uninstalling packages')
    state = processing_branch(
        packages, repo, state, extras=extras, is_force=is_force,
        commit_message=commit_message, username=username, user_email=user_email
    )
    if state.status() == ChangeStatus.no_change:
        logger.info(f'No changes in {repo.git.active_branch.name} branch. Skip create/merge merge request')
        return state
    logger.info(':: Creating merge request')
    mr = create_mr(
        repo.gitlab, repo.git.active_branch.name, branch, title=commit_message,
        description=f'Installed package(s): {get_packages_str(packages)}'
    )
    logger.info(':: Merging merge request')
    merge_mr(repo.gitlab, mr)
    return state


def processing_branch(
        packages: tuple[PackageLocalData, ...], repo: RepositoryInfo, state: StateFile, *,
        extras: tuple[ExtraParam, ...], commit_message: str, username: str, user_email: str,
        is_force: bool
) -> StateFile:
    state = uninstall_packages(packages, repo, state, extras=extras, is_force=is_force)

    if state.status() == ChangeStatus.changed:
        logger.debug(f'Updated state in {repo.git.active_branch.name} branch:\n{state}')
        directory = Path(repo.git.working_dir)
        state.save(directory)

        push_installed_packages(repo, commit_message, name=username, email=user_email)
        branch_url = repo.gitlab.branches.get(str(repo.git.active_branch)).web_url
        logger.success(f'Changes have been pushed to {repo.git.active_branch.name} branch (url: {branch_url})')
    else:
        logger.info(f'No changes in {repo.git.active_branch.name} branch. Skip push')
    return state


def uninstall_packages(
        packages: tuple[PackageLocalData, ...], repo: RepositoryInfo, state: StateFile, *,
        extras: tuple[ExtraParam, ...], is_force: bool
) -> StateFile:
    if not state.get_packages():
        logger.info('No packages installed')
        return state
    created_cicd_variables = state.get_all_created_cicd_variables()
    for i, package in enumerate(packages, 1):
        logger.info(f"({i}/{len(packages)}) Uninstalling {package['name']}@{package['version']} package")
        package_state = state.get_package(package, for_delete=True)
        if package_state is None:
            logger.info(f"{package['name']}@{package['version']} package is not installed")
            continue
        try:
            delete(
                package, repo, package_state,
                created_cicd_variables=created_cicd_variables, extras=extras, is_force=is_force
            )
            state.delete_package(package_state)
        except Exception:
            logger.critical(f"An error came out at one stage of uninstallation. "
                            f"Uninstallation {package['name']}@{package['version']} aborted")
            raise UninstallationError('Uninstallation failed in one of the steps. Please check log above')
    return state
